# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import random
import unittest

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import test_config
from azure.cosmos import DatabaseProxy
from azure.cosmos._change_feed.feed_range_internal import FeedRangeInternalEpk
from azure.cosmos._routing.routing_range import Range

COLLECTION = "created_collection"
DATABASE = "created_db"
@pytest.fixture(scope="class")
def setup():
    if (TestSessionTokenHelpers.masterKey == '[YOUR_KEY_HERE]' or
            TestSessionTokenHelpers.host == '[YOUR_ENDPOINT_HERE]'):
        raise Exception(
            "You must specify your Azure Cosmos account values for "
            "'masterKey' and 'host' at the top of this class to run the "
            "tests.")
    test_client = cosmos_client.CosmosClient(TestSessionTokenHelpers.host, test_config.TestConfig.masterKey),
    created_db = test_client[0].get_database_client(TestSessionTokenHelpers.TEST_DATABASE_ID)
    return {
        DATABASE: created_db,
        COLLECTION: created_db.get_container_client(TestSessionTokenHelpers.TEST_COLLECTION_ID)
    }

def create_split_ranges():
    return [        # split with two children
                   ([(("AA", "DD"), "0:1#51#3=52"), (("AA", "BB"),"1:1#55#3=52"), (("BB", "DD"),"2:1#54#3=52")],
                    ("AA", "DD"), "1:1#55#3=52,2:1#54#3=52"),
                    # same range different partition key range ids
                   ([(("AA", "DD"), "1:1#51#3=52"), (("AA", "DD"),"0:1#55#3=52")],
                    ("AA", "DD"), "0:1#55#3=52"),
                    # split with one child
                   ([(("AA", "DD"), "0:1#51#3=52"), (("AA", "BB"),"1:1#55#3=52")],
                    ("AA", "DD"), "0:1#51#3=52,1:1#55#3=52"),
                    # Highest GLSN, which is 55 in this         # cspell:disable-line
                    # ex. "1:1#55#3=52", is the one that will be returned because
                    # it is higher than all of the other feed range contained in the same
                    # range
                   ([(("AA", "DD"), "0:1#42#3=52"), (("AA", "BB"), "1:1#51#3=52"),
                    (("BB", "CC"),"1:1#53#3=52"), (("CC", "DD"),"1:1#55#3=52")],
                    ("AA", "DD"), "1:1#55#3=52"),
                    # Highest GLSN, which is 60 in this            # cspell:disable-line
                    # ex. "1:1#60#3=52", is the one that will be returned because
                    # it is higher than all of the other feed range contained in the same
                    # range with some of them overlapping with each other
                   ([(("AA", "DD"), "0:1#60#3=52"), (("AA", "BB"), "1:1#51#3=52"),
                    (("BB", "CC"),"1:1#53#3=52"), (("CC", "DD"),"1:1#55#3=52")],
                    ("AA", "DD"), "0:1#60#3=52"),
                    # AA-DD can be created from the other ranges
                    # but the GLSN's are not all larger than the one  # cspell:disable-line
                    # in the AA-DD range so we just compound as cannot make
                    # conclusions in this case
                   ([(("AA", "DD"), "0:1#60#3=52"), (("AA", "BB"), "1:1#51#3=52"),
                    (("BB", "CC"),"1:1#66#3=52"), (("CC", "DD"),"1:1#55#3=52")],
                    ("AA", "DD"), "0:1#60#3=52,1:1#66#3=52"),
                    # merge with one child
                   ([(("AA", "DD"), "3:1#55#3=52"), (("AA", "BB"),"1:1#51#3=52")],
                    ("AA", "DD"), "3:1#55#3=52"),
                    # merge with two children
                   ([(("AA", "DD"), "3:1#55#3=52"), (("AA", "BB"),"1:1#51#3=52"), (("BB", "DD"),"2:1#54#3=52")],
                    ("AA", "DD"), "3:1#55#3=52"),
                    # compound session token
                   ([(("AA", "DD"), "2:1#54#3=52,1:1#55#3=52"), (("AA", "BB"),"0:1#51#3=52")],
                    ("AA", "BB"), "2:1#54#3=52,1:1#55#3=52,0:1#51#3=52"),
                    # several compound session token with one range
                   ([(("AA", "DD"), "2:1#57#3=52,1:1#57#3=52"), (("AA", "DD"),"2:1#56#3=52,1:1#58#3=52")],
                    ("AA", "DD"), "2:1#57#3=52,1:1#58#3=52"),
                    # overlapping ranges
                   ([(("AA", "CC"), "0:1#54#3=52"), (("BB", "FF"),"2:1#51#3=52")],
                    ("AA", "EE"), "0:1#54#3=52,2:1#51#3=52"),
                    # different version numbers
                   ([(("AA", "BB"), "0:1#54#3=52"), (("AA", "BB"),"0:2#57#3=53")],
                    ("AA", "BB"), "0:2#57#3=53"),
                    # mixed scenarios
                   ([(("AA", "DD"), "3:1#60#3=53"), (("AA", "BB"), "1:1#54#3=52"), (("AA", "BB"), "1:1#52#3=53"),
                     (("BB", "CC"),"1:1#53#3=52"), (("BB", "CC"),"6:1#70#3=55,4:1#90#3=52"),
                     (("CC", "DD"),"1:1#55#3=52")], ("AA", "DD"), "3:1#60#3=53,6:1#70#3=55,4:1#90#3=52")
                  ]

@pytest.mark.cosmosEmulator
@pytest.mark.unittest
@pytest.mark.usefixtures("setup")
class TestSessionTokenHelpers:
    """Test for session token helpers"""

    created_db: DatabaseProxy = None
    client: cosmos_client.CosmosClient = None
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    configs = test_config.TestConfig
    TEST_DATABASE_ID = configs.TEST_DATABASE_ID
    TEST_COLLECTION_ID = configs.TEST_SINGLE_PARTITION_CONTAINER_ID

    def test_get_session_token_update(self, setup):
        feed_range = FeedRangeInternalEpk(
            Range("AA", "BB", True, False)).to_dict()
        session_token = "0:1#54#3=50"
        feed_ranges_and_session_tokens = [(feed_range, session_token)]
        session_token = "0:1#51#3=52"
        feed_ranges_and_session_tokens.append((feed_range, session_token))
        session_token = setup[COLLECTION].get_latest_session_token(feed_ranges_and_session_tokens, feed_range)
        assert session_token == "0:1#54#3=52"

    def test_many_session_tokens_update_same_range(self, setup):
        feed_range = FeedRangeInternalEpk(
            Range("AA", "BB", True, False)).to_dict()
        feed_ranges_and_session_tokens = []
        for i in range(1000):
            session_token = "0:1#" + str(random.randint(1, 100)) + "#3=" + str(random.randint(1, 100))
            feed_ranges_and_session_tokens.append((feed_range, session_token))
        session_token = "0:1#101#3=101"
        feed_ranges_and_session_tokens.append((feed_range, session_token))
        updated_session_token = setup["created_collection"].get_latest_session_token(feed_ranges_and_session_tokens,
                                                                                     feed_range)
        assert updated_session_token == session_token

    def test_many_session_tokens_update(self, setup):
        feed_range = FeedRangeInternalEpk(
            Range("AA", "BB", True, False)).to_dict()
        feed_ranges_and_session_tokens = []
        for i in range(1000):
            session_token = "0:1#" + str(random.randint(1, 100)) + "#3=" + str(random.randint(1, 100))
            feed_ranges_and_session_tokens.append((feed_range, session_token))

        # adding irrelevant feed ranges
        feed_range1 = FeedRangeInternalEpk(
            Range("CC", "FF", True, False)).to_dict()
        feed_range2 = FeedRangeInternalEpk(
            Range("00", "55", True, False)).to_dict()
        for i in range(1000):
            session_token = "0:1#" + str(random.randint(1, 100)) + "#3=" + str(random.randint(1, 100))
            if i % 2 == 0:
                feed_ranges_and_session_tokens.append((feed_range1, session_token))
            else:
                feed_ranges_and_session_tokens.append((feed_range2, session_token))
        session_token = "0:1#101#3=101"
        feed_ranges_and_session_tokens.append((feed_range, session_token))
        updated_session_token = setup["created_collection"].get_latest_session_token(feed_ranges_and_session_tokens,
                                                                                     feed_range)
        assert updated_session_token == session_token

    @pytest.mark.parametrize("split_ranges, target_feed_range, expected_session_token", create_split_ranges())
    def test_simulated_splits_merges(self, setup, split_ranges, target_feed_range, expected_session_token):
        actual_split_ranges = []
        for feed_range, session_token in split_ranges:
            actual_split_ranges.append((FeedRangeInternalEpk(Range(feed_range[0], feed_range[1],
                                                True, False)).to_dict(), session_token))
        target_feed_range = FeedRangeInternalEpk(Range(target_feed_range[0], target_feed_range[1][1],
                                               True, False)).to_dict()
        updated_session_token = setup[COLLECTION].get_latest_session_token(actual_split_ranges, target_feed_range)
        assert updated_session_token == expected_session_token

    def test_invalid_feed_range(self, setup):
        feed_range = FeedRangeInternalEpk(
            Range("AA", "BB", True, False)).to_dict()
        session_token = "0:1#54#3=50"
        feed_ranges_and_session_tokens = [(feed_range, session_token)]
        with pytest.raises(ValueError, match='There were no overlapping feed ranges with the target.'):
            setup["created_collection"].get_latest_session_token(feed_ranges_and_session_tokens,
                                                                 FeedRangeInternalEpk(Range(
                                                                      "CC",
                                                                      "FF",
                                                                      True,
                                                                      False)).to_dict())

if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal Helper functions in the Azure Cosmos database service.
"""

import platform
import re
import base64
import json
import time
from typing import Any, Optional, Tuple

from ._version import VERSION


def get_user_agent(suffix: Optional[str] = None) -> str:
    os_name = safe_user_agent_header(platform.platform())
    python_version = safe_user_agent_header(platform.python_version())
    user_agent = "azsdk-python-cosmos/{} Python/{} ({})".format(VERSION, python_version, os_name)
    if suffix:
        user_agent += f" {suffix}"
    return user_agent

def get_user_agent_async(suffix: Optional[str] = None) -> str:
    os_name = safe_user_agent_header(platform.platform())
    python_version = safe_user_agent_header(platform.python_version())
    user_agent = "azsdk-python-cosmos-async/{} Python/{} ({})".format(VERSION, python_version, os_name)
    if suffix:
        user_agent += f" {suffix}"
    return user_agent


def safe_user_agent_header(s: Optional[str] = None) -> str:
    if s is None:
        s = "unknown"
    # remove all white spaces
    s = re.sub(r"\s+", "", s)
    if not s:
        s = "unknown"
    return s


def get_index_metrics_info(delimited_string: Optional[str] = None) -> dict[str, Any]:
    if delimited_string is None:
        return {}
    try:
        # Decode the base64 string to bytes
        bytes_string = base64.b64decode(delimited_string)
        # Decode the bytes to a string using UTF-8 encoding
        decoded_string = bytes_string.decode('utf-8')

        # Python's json.loads method is used for deserialization
        result = json.loads(decoded_string) or {}
        return result
    except (json.JSONDecodeError, ValueError):
        return {}

def current_time_millis() -> int:
    return int(round(time.time() * 1000))

def add_args_to_kwargs(
        arg_names: list[str],
        args: Tuple[Any, ...],
        kwargs: dict[str, Any]
    ) -> None:
    """Add positional arguments(args) to keyword argument dictionary(kwargs) using names in arg_names as keys.
    To be backward-compatible, some expected positional arguments has to be allowed. This method will verify number of
    maximum positional arguments and add them to the keyword argument dictionary(kwargs)

    :param list[str] arg_names: The names of positional arguments.
    :param Tuple[Any, ...] args: The tuple of positional arguments.
    :param dict[str, Any] kwargs: The dictionary of keyword arguments as reference. This dictionary will be updated.
    """

    if len(args) > len(arg_names):
        raise ValueError(f"Positional argument is out of range. Expected {len(arg_names)} arguments, "
                         f"but got {len(args)} instead. Please review argument list in API documentation.")

    for name, arg in zip(arg_names, args):
        if name in kwargs:
            raise ValueError(f"{name} cannot be used as positional and keyword argument at the same time.")
        kwargs[name] = arg


def format_list_with_and(items: list[str]) -> str:
    """Format a list of items into a string with commas and 'and' for the last item.

    :param list[str] items: The list of items to format.
    :return: A formatted string with items separated by commas and 'and' before the last item.
    :rtype: str
    """
    formatted_items = ""
    quoted = [f"'{item}'" for item in items]
    if len(quoted) > 2:
        formatted_items = ", ".join(quoted[:-1]) + ", and " + quoted[-1]
    elif len(quoted) == 2:
        formatted_items = " and ".join(quoted)
    elif quoted:
        formatted_items = quoted[0]
    return formatted_items

def verify_exclusive_arguments(
        exclusive_keys: list[str],
        **kwargs: dict[str, Any]) -> None:
    """Verify if exclusive arguments are present in kwargs.
    For some Cosmos SDK APIs, some arguments are exclusive, or cannot be used at the same time. This method will verify
    that and raise an error if exclusive arguments are present.

    :param list[str] exclusive_keys: The names of exclusive arguments.
    """
    keys_in_kwargs = [key for key in exclusive_keys if key in kwargs and kwargs[key] is not None]

    if len(keys_in_kwargs) > 1:
        raise ValueError(f"{format_list_with_and(keys_in_kwargs)} are exclusive parameters, "
                         f"please only set one of them.")

def valid_key_value_exist(
        kwargs: dict[str, Any],
        key: str,
        invalid_value: Any = None) -> bool:
    """Check if a valid key and value exists in kwargs. It always checks if the value is not None and it will remove
    from the kwargs the None value.

    :param dict[str, Any] kwargs: The dictionary of keyword arguments.
    :param str key: The key to check.
    :param Any invalid_value: The value that is considered invalid. Default is None.
    :return: True if the key exists and its value is not None, False otherwise.
    :rtype: bool
    """
    if key in kwargs and kwargs[key] is None:
        kwargs.pop(key)
        return False

    return key in kwargs and kwargs[key] is not invalid_value



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Range partition resolver implementation in the Azure Cosmos database service.
"""

from . import _range as prange


class RangePartitionResolver(object):
    """RangePartitionResolver implements partitioning based on the ranges,
    allowing you to distribute requests and data across a number of partitions.
    """

    def __init__(self, partition_key_extractor, partition_map):
        """
        :param lambda partition_key_extractor:
            Returning the partition key from the document passed.
        :param dict partition_map:
            The dictionary of ranges mapped to their associated collection
        """
        if partition_key_extractor is None:
            raise ValueError("partition_key_extractor is None.")
        if partition_map is None:
            raise ValueError("partition_map is None.")

        self.partition_key_extractor = partition_key_extractor
        self.partition_map = partition_map

    def ResolveForCreate(self, document):
        """Resolves the collection for creating the document based on the partition key.

        :param dict document: The document to be created.
        :return: Collection Self link or Name based link which should handle the Create operation.
        :rtype: str
        """
        if document is None:
            raise ValueError("document is None.")

        partition_key = self.partition_key_extractor(document)
        containing_range = self._GetContainingRange(partition_key)

        if containing_range is None:
            raise ValueError("A containing range for " + str(partition_key) + " doesn't exist in the partition map.")

        return self.partition_map.get(containing_range)

    def ResolveForRead(self, partition_key):
        """Resolves the collection for reading/querying the documents based on the partition key.

        :param str partition_key: The partition key to be used.
        :return: Collection Self link(s) or Name based link(s) which should handle the Read operation.
        :rtype: list
        """
        intersecting_ranges = self._GetIntersectingRanges(partition_key)

        collection_links = []
        for keyrange in intersecting_ranges:
            collection_links.append(self.partition_map.get(keyrange))

        return collection_links

    def _GetContainingRange(self, partition_key):
        """Get the containing range based on the partition key.
        :param str partition_key: The partition key to be used.
        :returns: The containing key range.
        :rtype: str
        """
        for keyrange in self.partition_map.keys():
            if keyrange.Contains(partition_key):
                return keyrange

        return None

    def _GetIntersectingRanges(self, partition_key):
        """Get the intersecting ranges based on the partition key.
        :param str partition_key: The partition key to be used.
        :returns: the set of intersecting ranges for the partition key.
        :rtype: set
        """
        partitionkey_ranges = set()
        intersecting_ranges = set()

        if partition_key is None:
            return list(self.partition_map.keys())

        if isinstance(partition_key, prange.Range):
            partitionkey_ranges.add(partition_key)
        elif isinstance(partition_key, list):
            for key in partition_key:
                if key is None:
                    return list(self.partition_map.keys())
                if isinstance(key, prange.Range):
                    partitionkey_ranges.add(key)
                else:
                    partitionkey_ranges.add(prange.Range(key, key))
        else:
            partitionkey_ranges.add(prange.Range(partition_key, partition_key))

        for partitionKeyRange in partitionkey_ranges:
            for keyrange in self.partition_map.keys():
                if keyrange.Intersect(partitionKeyRange):
                    intersecting_ranges.add(keyrange)

        return intersecting_ranges



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for change feed composite continuation token in the Azure Cosmos
database service.
"""
from typing import Optional, Any

from azure.cosmos._routing.routing_range import Range


class CompositeContinuationToken:
    token_property_name = "token"
    feed_range_property_name = "range"

    def __init__(self, feed_range: Range, token: Optional[str] = None) -> None:
        if feed_range is None:
            raise ValueError("Missing required parameter feed_range")

        self._token = token
        self._feed_range = feed_range

    def to_dict(self) -> dict[str, Any]:
        return {
            self.token_property_name: self._token,
            self.feed_range_property_name: self.feed_range.to_dict()
        }

    @property
    def feed_range(self) -> Range:
        return self._feed_range

    @property
    def token(self) -> Optional[str]:
        return self._token

    def update_token(self, etag) -> None:
        self._token = etag

    @classmethod
    def from_json(cls, data) -> 'CompositeContinuationToken':
        token = data.get(cls.token_property_name)
        if token is None:
            raise ValueError(f"Invalid composite token [Missing {cls.token_property_name}]")

        feed_range_data = data.get(cls.feed_range_property_name)
        if feed_range_data is None:
            raise ValueError(f"Invalid composite token [Missing {cls.feed_range_property_name}]")

        feed_range = Range.ParseFromDict(feed_range_data)
        return cls(feed_range=feed_range, token=token)

    def __repr__(self):
        return f"CompositeContinuationToken(token={self.token}, range={self.feed_range})"



# -*- coding: utf-8 -*-
# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid
import pytest
from azure.core.exceptions import ServiceRequestError
from azure.cosmos.aio import CosmosClient, DatabaseProxy
import test_config
from azure.cosmos import PartitionKey
from _fault_injection_transport import FaultInjectionTransport
from azure.cosmos.documents import _OperationType, ConnectionPolicy


@pytest.mark.cosmosMultiRegion
class TestServiceRequestRetryPoliciesAsync(unittest.TestCase):
    """Test cases for the read_items API."""

    created_db: DatabaseProxy = None
    client: CosmosClient = None
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    configs = test_config.TestConfig
    TEST_DATABASE_ID = configs.TEST_DATABASE_ID

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.masterKey)
        self.database = self.client.get_database_client(self.TEST_DATABASE_ID)
        self.container = await self.database.create_container("service_request_mrr_test_" + str(uuid.uuid4()),
                                                              PartitionKey(path="/id"))

    async def asyncTearDown(self):
        await self.client.close()

    async def test_write_failover_to_global_with_service_request_error_async(self):
        # 1. Get write regions and ensure there are at least 2 for this test.
        endpoint_manager = self.client.client_connection._global_endpoint_manager
        db_account = await self.client._get_database_account()
        endpoint_manager.refresh_endpoint_list(db_account)

        write_locations_map = endpoint_manager.location_cache.account_locations_by_write_endpoints

        if len(write_locations_map) < 2:
            pytest.skip("This test requires an account with 2 or more write regions.")

        write_endpoints = list(write_locations_map.keys())
        write_locations = list(write_locations_map.values())

        account_name = self.host.split(".")[0].split("//")[1]
        region_to_fail_slug = write_endpoints[1].split('/')[2].replace(f"{account_name}-", "").split('.')[0]

        region_to_exclude = write_locations[0]  # Use the display name for ExcludedLocations

        # 2. Set up a client with one region excluded and a fault injection transport.
        policy = ConnectionPolicy()  # Create a new policy object for this test
        policy.EnableEndpointDiscovery = True
        policy.UseMultipleWriteLocations = True
        policy.ExcludedLocations = [region_to_exclude]
        fault_injection_transport = FaultInjectionTransport()

        async with CosmosClient(
                self.host,
                self.masterKey,
                connection_policy=policy,
                transport=fault_injection_transport,

        ) as client_with_faults:
            container_with_faults = client_with_faults.get_database_client(self.database.id).get_container_client(
                self.container.id)

        # 3. Configure fault injection to fail requests to the second write region with a ServiceRequestError.
        error_to_inject = ServiceRequestError(message="Simulated Service Request Error")

        def predicate(request):
            # Fail if it's a create item operation and the host matches the region we want to fail.
            is_create = FaultInjectionTransport.predicate_is_operation_type(request, _OperationType.Create)
            is_target_region = region_to_fail_slug in request.url
            return is_create and is_target_region

        def fault_action(_):
            raise error_to_inject

        fault_injection_transport.add_fault(predicate, fault_action)

        # 4. Execute a write operation. It should fail with ServiceRequestError as no regions are available.
        with self.assertRaises(ServiceRequestError) as context:
            await container_with_faults.create_item(body={'id': 'failover_test_id', 'pk': 'pk_value'})

        self.assertIn("Simulated Service Request Error", str(context.exception))


if __name__ == "__main__":
    unittest.main()




# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import os

settings = {
    'host': os.environ.get('ACCOUNT_HOST', '[YOUR ENDPOINT]'),
    'master_key': os.environ.get('ACCOUNT_KEY', '[YOUR KEY]'),
    'database_id': os.environ.get('COSMOS_DATABASE', '[YOUR DATABASE]'),
    'container_id': os.environ.get('COSMOS_CONTAINER', '[YOUR CONTAINER]'),
    'tenant_id': os.environ.get('TENANT_ID', '[YOUR TENANT ID]'),
    'client_id': os.environ.get('CLIENT_ID', '[YOUR CLIENT ID]'),
    'client_secret': os.environ.get('CLIENT_SECRET', '[YOUR CLIENT SECRET]'),
    'container_mh_id': os.environ.get('COSMOS_CONTAINER_MH', '[YOUR MH CONTAINER]'),
}



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import asyncio
import os
import unittest
import uuid
from typing import Any

import pytest
from azure.core.pipeline.transport._aiohttp import AioHttpTransport
from azure.core.exceptions import ServiceResponseError

import test_config
from azure.cosmos import _location_cache, _partition_health_tracker
from azure.cosmos.aio import CosmosClient
from azure.cosmos.exceptions import CosmosHttpResponseError
from _fault_injection_transport_async import FaultInjectionTransportAsync
from test_per_partition_circuit_breaker_mm import create_doc, read_operations_and_errors, \
    write_operations_and_errors, operations, REGION_1, REGION_2, CHANGE_FEED, CHANGE_FEED_PK, CHANGE_FEED_EPK, READ, \
    CREATE, READ_ALL_ITEMS, DELETE_ALL_ITEMS_BY_PARTITION_KEY, QUERY, QUERY_PK, BATCH, UPSERT, REPLACE, PATCH, DELETE, \
    PK_VALUE, validate_unhealthy_partitions, validate_response_uri
from test_per_partition_circuit_breaker_mm import validate_stats

COLLECTION = "created_collection"

async def perform_write_operation(operation, container, fault_injection_container, doc_id, pk, expected_uri):
    doc = {'id': doc_id,
           'pk': pk,
           'name': 'sample document',
           'key': 'value'}
    if operation == CREATE:
        resp = await fault_injection_container.create_item(body=doc)
    elif operation == UPSERT:
        resp = await fault_injection_container.upsert_item(body=doc)
    elif operation == REPLACE:
        await container.create_item(body=doc)
        new_doc = {'id': doc_id,
                   'pk': pk,
                   'name': 'sample document' + str(uuid),
                   'key': 'value'}
        await asyncio.sleep(1)
        resp = await fault_injection_container.replace_item(item=doc['id'], body=new_doc)
    elif operation == DELETE:
        await container.create_item(body=doc)
        await asyncio.sleep(1)
        resp = await fault_injection_container.delete_item(item=doc['id'], partition_key=doc['pk'])
    elif operation == PATCH:
        await container.create_item(body=doc)
        await asyncio.sleep(1)
        operations = [{"op": "incr", "path": "/company", "value": 3}]
        resp = await fault_injection_container.patch_item(item=doc['id'], partition_key=doc['pk'], patch_operations=operations)
    elif operation == BATCH:
        batch_operations = [
            ("create", (doc, )),
            ("upsert", (doc,)),
            ("upsert", (doc,)),
            ("upsert", (doc,)),
        ]
        resp = await fault_injection_container.execute_item_batch(batch_operations, partition_key=doc['pk'])
    # this will need to be emulator only
    elif operation == DELETE_ALL_ITEMS_BY_PARTITION_KEY:
        await container.create_item(body=doc)
        resp = await fault_injection_container.delete_all_items_by_partition_key(pk)
    if resp:
        validate_response_uri(resp, expected_uri)

async def perform_read_operation(operation, container, doc_id, pk, expected_uri):
    if operation == READ:
        read_resp = await container.read_item(item=doc_id, partition_key=pk)
        request = read_resp.get_response_headers()["_request"]
        # Validate the response comes from "Read Region" (the most preferred read-only region)
        assert request.url.startswith(expected_uri)
    elif operation == QUERY_PK:
        # partition key filtered query
        query = "SELECT * FROM c WHERE c.id = @id AND c.pk = @pk"
        parameters = [{"name": "@id", "value": doc_id}, {"name": "@pk", "value": pk}]
        async for item in container.query_items(query=query, partition_key=pk, parameters=parameters):
            assert item['id'] == doc_id
        # need to do query with no pk and with feed range
    elif operation == QUERY:
        # cross partition query
        query = "SELECT * FROM c WHERE c.id = @id"
        async for item in container.query_items(query=query):
            assert item['id'] == doc_id
    elif operation == CHANGE_FEED:
        async for _ in container.query_items_change_feed():
            pass
    elif operation == CHANGE_FEED_PK:
        # partition key filtered change feed
        async for _ in container.query_items_change_feed(partition_key=pk):
            pass
    elif operation == CHANGE_FEED_EPK:
        # partition key filtered by feed range
        feed_range = await container.feed_range_from_partition_key(partition_key=pk)
        async for _ in container.query_items_change_feed(feed_range=feed_range):
            pass
    elif operation == READ_ALL_ITEMS:
        async for _ in container.read_all_items():
            pass

async def cleanup_method(initialized_objects: list[dict[str, Any]]):
    for obj in initialized_objects:
        method_client: CosmosClient = obj["client"]
        await method_client.close()

@pytest.mark.cosmosCircuitBreaker
@pytest.mark.asyncio
class TestPerPartitionCircuitBreakerMMAsync:
    host = test_config.TestConfig.host
    master_key = test_config.TestConfig.masterKey
    TEST_DATABASE_ID = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_MULTI_PARTITION_ID = test_config.TestConfig.TEST_MULTI_PARTITION_CONTAINER_ID

    async def setup_method_with_custom_transport(self, custom_transport: AioHttpTransport, default_endpoint=host, **kwargs):
        container_id = kwargs.pop("container_id", None)
        if not container_id:
            container_id = self.TEST_CONTAINER_MULTI_PARTITION_ID
        client = CosmosClient(default_endpoint, self.master_key,
                              preferred_locations=[REGION_1, REGION_2],
                              multiple_write_locations=True,
                              transport=custom_transport, **kwargs)
        await client.__aenter__()
        db = client.get_database_client(self.TEST_DATABASE_ID)
        container = db.get_container_client(container_id)
        return {"client": client, "db": db, "col": container}

    @pytest.mark.parametrize("write_operation, error", write_operations_and_errors())
    async def test_write_consecutive_failure_threshold_async(self, write_operation, error):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0,
            error
        ))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

        with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
            await perform_write_operation(write_operation,
                                          container,
                                          fault_injection_container,
                                          str(uuid.uuid4()),
                                          PK_VALUE,
                                          expected_uri)
        assert exc_info.value == error

        validate_unhealthy_partitions(global_endpoint_manager, 0)

        # writes should fail but still be tracked
        for i in range(4):
            with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
                await perform_write_operation(write_operation,
                                              container,
                                              fault_injection_container,
                                              str(uuid.uuid4()),
                                              PK_VALUE,
                                              expected_uri)
            assert exc_info.value == error

        # writes should now succeed because going to the other region
        await perform_write_operation(write_operation,
                                      container,
                                      fault_injection_container,
                                      str(uuid.uuid4()),
                                      PK_VALUE,
                                      expected_uri)

        validate_unhealthy_partitions(global_endpoint_manager, 1)
        # remove faults and reduce initial recover time and perform a write
        original_unavailable_time = _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS
        _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = 1
        custom_transport.faults = []
        try:
            await perform_write_operation(write_operation,
                                          container,
                                          fault_injection_container,
                                          str(uuid.uuid4()),
                                          PK_VALUE,
                                          uri_down)
        finally:
            _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = original_unavailable_time
        validate_unhealthy_partitions(global_endpoint_manager, 0)
        await cleanup_method([custom_setup, setup])

    async def setup_info(self, error, **kwargs):
        expected_uri = _location_cache.LocationCache.GetLocationalEndpoint(self.host, REGION_2)
        uri_down = _location_cache.LocationCache.GetLocationalEndpoint(self.host, REGION_1)
        custom_transport = FaultInjectionTransportAsync()
        # two documents targeted to same partition, one will always fail and the other will succeed
        doc = create_doc()
        predicate = lambda r: (FaultInjectionTransportAsync.predicate_is_document_operation(r) and
                               FaultInjectionTransportAsync.predicate_targets_region(r, uri_down))
        custom_transport.add_fault(predicate,
                                   error)
        custom_setup = await self.setup_method_with_custom_transport(custom_transport, default_endpoint=self.host, **kwargs)
        setup = await self.setup_method_with_custom_transport(None, default_endpoint=self.host, **kwargs)
        return setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate


    @pytest.mark.cosmosCircuitBreakerMultiRegion
    @pytest.mark.parametrize("read_operation, error", read_operations_and_errors())
    async def test_read_consecutive_failure_threshold_async(self, read_operation, error):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0,
            error
        ))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']

        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        
        # create a document to read
        await container.create_item(body=doc)

        # reads should fail over and only the relevant partition should be marked as unavailable
        await perform_read_operation(read_operation,
                                     fault_injection_container,
                                     doc['id'],
                                     doc['pk'],
                                     expected_uri)
        # partition should not have been marked unavailable after one error
        validate_unhealthy_partitions(global_endpoint_manager, 0)

        for i in range(10):
            await perform_read_operation(read_operation,
                                         fault_injection_container,
                                         doc['id'],
                                         doc['pk'],
                                         expected_uri)

        # the partition should have been marked as unavailable after breaking read threshold
        if read_operation in (CHANGE_FEED, QUERY, READ_ALL_ITEMS):
            # these operations are cross partition so they would mark both partitions as unavailable
            expected_unhealthy_partitions = 5
        else:
            expected_unhealthy_partitions = 1
        validate_unhealthy_partitions(global_endpoint_manager, expected_unhealthy_partitions)
        # remove faults and reduce initial recover time and perform a read
        original_unavailable_time = _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS
        _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = 1
        custom_transport.faults = []
        try:
            await perform_read_operation(read_operation,
                                         fault_injection_container,
                                         doc['id'],
                                         doc['pk'],
                                         uri_down)
        finally:
            _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = original_unavailable_time
        validate_unhealthy_partitions(global_endpoint_manager, 0)
        await cleanup_method([custom_setup, setup])

    @pytest.mark.parametrize("write_operation, error", write_operations_and_errors())
    async def test_write_failure_rate_threshold_async(self, write_operation, error):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0,
            error
        ))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        # lower minimum requests for testing
        _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 10
        os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "80"
        try:
            # writes should fail but still be tracked and mark unavailable a partition after crossing threshold
            for i in range(10):
                validate_unhealthy_partitions(global_endpoint_manager, 0)
                if i == 4 or i == 8:
                    # perform some successful creates to reset consecutive counter
                    # remove faults and perform a write
                    custom_transport.faults = []
                    await fault_injection_container.upsert_item(body=doc)
                    custom_transport.add_fault(predicate,
                                               lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
                                                   0,
                                                   error
                                               )))
                else:
                    with pytest.raises((CosmosHttpResponseError, ServiceResponseError)) as exc_info:
                        await perform_write_operation(write_operation,
                                                      container,
                                                      fault_injection_container,
                                                      str(uuid.uuid4()),
                                                      PK_VALUE,
                                                      expected_uri)
                    assert exc_info.value == error

            validate_unhealthy_partitions(global_endpoint_manager, 1)

        finally:
            os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "90"
            # restore minimum requests
            _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 100
        await cleanup_method([custom_setup, setup])

    @pytest.mark.cosmosCircuitBreakerMultiRegion
    @pytest.mark.parametrize("read_operation, error", read_operations_and_errors())
    async def test_read_failure_rate_threshold_async(self, read_operation, error):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0,
            error
        ))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        await container.upsert_item(body=doc)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        # lower minimum requests for testing
        _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 8
        os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "80"
        try:
            if isinstance(error, ServiceResponseError):
                # service response error retries in region 3 additional times before failing over
                num_operations = 2
            else:
                num_operations = 8
            for i in range(num_operations):
                validate_unhealthy_partitions(global_endpoint_manager, 0)
                # read will fail and retry in other region
                await perform_read_operation(read_operation,
                                             fault_injection_container,
                                             doc['id'],
                                             PK_VALUE,
                                             expected_uri)
            if read_operation in (CHANGE_FEED, QUERY, READ_ALL_ITEMS):
                # these operations are cross partition so they would mark both partitions as unavailable
                expected_unhealthy_partitions = 5
            else:
                expected_unhealthy_partitions = 1

            validate_unhealthy_partitions(global_endpoint_manager, expected_unhealthy_partitions)
        finally:
            os.environ["AZURE_COSMOS_FAILURE_PERCENTAGE_TOLERATED"] = "90"
            # restore minimum requests
            _partition_health_tracker.MINIMUM_REQUESTS_FOR_FAILURE_RATE = 100
        await cleanup_method([custom_setup, setup])

    async def test_stat_reset_async(self):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0,
            CosmosHttpResponseError(
                status_code=503,
                message="Some injected error.")
        ))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = \
            await self.setup_info(error_lambda, container_id=test_config.TestConfig.TEST_SINGLE_PARTITION_CONTAINER_ID)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        await container.upsert_item(body=doc)
        await asyncio.sleep(1)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager
        # lower refresh interval for testing
        _partition_health_tracker.REFRESH_INTERVAL_MS = 10 * 1000
        try:
            for i in range(2):
                validate_unhealthy_partitions(global_endpoint_manager, 0)
                # read will fail and retry in other region
                await perform_read_operation(READ,
                                             fault_injection_container,
                                             doc['id'],
                                             PK_VALUE,
                                             expected_uri)
                try:
                    await perform_write_operation(CREATE,
                                                  container,
                                                  fault_injection_container,
                                                  str(uuid.uuid4()),
                                                  PK_VALUE,
                                                  expected_uri)
                except CosmosHttpResponseError as e:
                    assert e.status_code == 503
            validate_unhealthy_partitions(global_endpoint_manager, 0)
            validate_stats(global_endpoint_manager, 2, 2, 2, 2, 0, 0)
            await asyncio.sleep(25)
            await perform_read_operation(READ,
                                         fault_injection_container,
                                         doc['id'],
                                         PK_VALUE,
                                         expected_uri)

            validate_stats(global_endpoint_manager, 2, 3, 1, 0, 0, 0)
        finally:
            _partition_health_tracker.REFRESH_INTERVAL_MS = 60 * 1000
            await cleanup_method([custom_setup, setup])

    @pytest.mark.parametrize("read_operation, write_operation", operations())
    async def test_service_request_error_async(self, read_operation, write_operation):
        # the region should be tried 4 times before failing over and mark the partition as unavailable
        # the region should not be marked as unavailable
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_region_down())
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        container = setup['col']
        fault_injection_container = custom_setup['col']
        await container.upsert_item(body=doc)
        await perform_read_operation(read_operation,
                                     fault_injection_container,
                                     doc['id'],
                                     PK_VALUE,
                                     expected_uri)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

        validate_unhealthy_partitions(global_endpoint_manager, 0)
        assert len(global_endpoint_manager.location_cache.location_unavailability_info_by_endpoint) == 1

        # recover partition
        # remove faults and reduce initial recover time and perform a read
        original_unavailable_time = _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS
        _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = 1
        custom_transport.faults = []
        try:
            await perform_read_operation(read_operation,
                                         fault_injection_container,
                                         doc['id'],
                                         PK_VALUE,
                                         expected_uri)
        finally:
            _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = original_unavailable_time
        validate_unhealthy_partitions(global_endpoint_manager, 0)
        # per partition circuit breaker should not regress connection timeouts marking the region as unavailable
        assert len(global_endpoint_manager.location_cache.location_unavailability_info_by_endpoint) == 1

        custom_transport.add_fault(predicate,
                                   lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_region_down()))

        await perform_write_operation(write_operation,
                                      container,
                                      fault_injection_container,
                                      str(uuid.uuid4()),
                                      PK_VALUE,
                                      expected_uri)
        global_endpoint_manager = fault_injection_container.client_connection._global_endpoint_manager

        validate_unhealthy_partitions(global_endpoint_manager, 0)
        assert len(global_endpoint_manager.location_cache.location_unavailability_info_by_endpoint) == 1


        await cleanup_method([custom_setup, setup])


    # send 15 write concurrent requests when trying to recover
    # verify that only one failed
    async def test_recovering_only_fails_one_requests_async(self):
        error_lambda = lambda r: asyncio.create_task(FaultInjectionTransportAsync.error_after_delay(
            0, CosmosHttpResponseError(
            status_code=502,
            message="Some envoy error.")))
        setup, doc, expected_uri, uri_down, custom_setup, custom_transport, predicate = await self.setup_info(error_lambda)
        fault_injection_container = custom_setup['col']
        for i in range(5):
            with pytest.raises(CosmosHttpResponseError):
                await fault_injection_container.create_item(body=doc)


        number_of_errors = 0

        async def concurrent_upsert():
            nonlocal number_of_errors
            doc = {'id': str(uuid.uuid4()),
                   'pk': PK_VALUE,
                   'name': 'sample document',
                   'key': 'value'}
            try:
                await fault_injection_container.upsert_item(doc)
            except CosmosHttpResponseError as e:
                number_of_errors += 1

        # attempt to recover partition
        original_unavailable_time = _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS
        _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = 1
        try:
            tasks = []
            for i in range(15):
                tasks.append(concurrent_upsert())
            await asyncio.gather(*tasks)
            assert number_of_errors == 1
        finally:
            _partition_health_tracker.INITIAL_UNAVAILABLE_TIME_MS = original_unavailable_time
            await cleanup_method([custom_setup, setup])

if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import asyncio
from typing import Any

from aiohttp import TCPConnector, ClientRequest, tracing, ClientTimeout, ClientConnectorError, client_proto


class ProxiedTCPConnector(TCPConnector):
    """
    A TCP connector that proxies all connections through a specified host and port.

    This class extends TrackedTCPConnector to override connection behavior,
    routing all traffic through a specified proxy server using non-encrypted http.
    """

    def __init__(
            self,
            *,
            proxy_host: str,
            proxy_port: int,
            **kwargs: Any,
    ) -> None:
        """
        Initialize the ProxiedTCPConnector.

        Args:
            proxy_host (str): The hostname of the proxy server.
            proxy_port (int): The port number of the proxy server.
            **kwargs: Additional keyword arguments passed to the base TCPConnector.
        """
        super().__init__(**kwargs)
        self.__proxy_host = proxy_host
        self.__proxy_port = proxy_port

    async def _create_direct_connection(
            self,
            req: ClientRequest,
            traces: list[tracing.Trace],
            timeout: ClientTimeout,
            *,
            client_error: type[Exception] = ClientConnectorError,
    ) -> tuple[asyncio.Transport, client_proto.ResponseHandler]:
        """Override host, port, and schema to use proxy"""
        req.url = (
            req.url.with_host(self.__proxy_host).with_port(self.__proxy_port).with_scheme("http")
        )
        return await super()._create_direct_connection(
            req,
            traces,
            timeout,
            client_error=client_error,
        )


# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import unittest
import uuid

import pytest

import test_config
from azure.cosmos import CosmosClient


@pytest.mark.cosmosEmulator
class TestRegionalRoutingContext(unittest.TestCase):
    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    TEST_DATABASE_ID = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_ID = test_config.TestConfig.TEST_SINGLE_PARTITION_CONTAINER_ID

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")
        cls.client = CosmosClient(cls.host, cls.masterKey)
        cls.created_database = cls.client.get_database_client(cls.TEST_DATABASE_ID)
        cls.created_container = cls.created_database.get_container_client(cls.TEST_CONTAINER_ID)

    def test_no_swaps_on_successful_request(self):

        original_read_endpoint = (self.client.client_connection._global_endpoint_manager
                                  .location_cache.get_read_regional_routing_context())
        self.created_container.create_item(body={"id": str(uuid.uuid4())})
        # Check for if there was a swap
        self.assertEqual(original_read_endpoint,
                         self.client.client_connection._global_endpoint_manager
                         .location_cache.get_read_regional_routing_context())
        self.assertEqual(original_read_endpoint,
                         self.client.client_connection._global_endpoint_manager
                         .location_cache.get_write_regional_routing_context())



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ._version import VERSION
from ._cosmos_responses import CosmosDict, CosmosList
from ._retry_utility import ConnectionRetryPolicy
from .container import ContainerProxy
from .cosmos_client import CosmosClient
from .database import DatabaseProxy
from .user import UserProxy
from .scripts import ScriptsProxy
from .offer import Offer
from .offer import ThroughputProperties
from .documents import (
    ConsistencyLevel,
    DataType,
    IndexKind,
    IndexingMode,
    PermissionMode,
    ProxyConfiguration,
    SSLConfiguration,
    TriggerOperation,
    TriggerType,
    DatabaseAccount,
)
from .partition_key import PartitionKey
from .permission import Permission

__all__ = (
    "CosmosClient",
    "DatabaseProxy",
    "ContainerProxy",
    "PartitionKey",
    "Permission",
    "ScriptsProxy",
    "UserProxy",
    "Offer",
    "DatabaseAccount",
    "ConsistencyLevel",
    "DataType",
    "IndexKind",
    "IndexingMode",
    "PermissionMode",
    "ProxyConfiguration",
    "SSLConfiguration",
    "TriggerOperation",
    "TriggerType",
    "ConnectionRetryPolicy",
    "ThroughputProperties",
    "CosmosDict",
    "CosmosList"
)
__version__ = VERSION



# -*- coding: utf-8 -*-
# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

"""Tests for the CosmosHttpLoggingPolicy."""
import logging
import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
from azure.cosmos import PartitionKey, ContainerProxy
from azure.cosmos.exceptions import CosmosHttpResponseError
import test_config
from _fault_injection_transport import FaultInjectionTransport
from test_fault_injection_transport import TestFaultInjectionTransport
from typing import List, Callable
from azure.core.rest import HttpRequest

try:
    from unittest.mock import Mock
except ImportError:  # python < 3.3
    from mock import Mock  # type: ignore


class FilterStatusCode(logging.Filter):
    def filter(self, record):
        if hasattr(record, 'status_code') and record.status_code >= 400:
            return True
        return False

CONFIG = test_config.TestConfig
L1 = "Location1"
L2 = "Location2"
L1_URL = test_config.TestConfig.local_host
L2_URL = L1_URL.replace("localhost", "127.0.0.1")
URL_TO_LOCATIONS = {
    L1_URL: L1,
    L2_URL: L2}


def create_logger(name: str, mock_handler: test_config.MockHandler, level: int = logging.INFO) -> logging.Logger:
    logger = logging.getLogger(name)
    logger.addHandler(mock_handler)
    logger.setLevel(level)

    return logger

def get_locations_list(msg: str) -> List[str]:
    msg = msg.replace(' ', '')
    msg = msg.replace('\'', '')
    # Find the substring between the first '[' and the last ']'
    start = msg.find('[') + 1
    end = msg.rfind(']')
    # Extract the substring and convert it to a list using ast.literal_eval
    msg = msg[start:end]
    return msg.split(',')

@pytest.mark.cosmosEmulator
class TestCosmosHttpLogger(unittest.TestCase):
    mock_handler_diagnostic = None
    mock_handler_default = None
    logger_diagnostic = None
    logger_default = None
    config = test_config.TestConfig
    host = config.host
    masterKey = config.masterKey
    connectionPolicy = config.connectionPolicy

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")
        cls.mock_handler_default = test_config.MockHandler()
        cls.mock_handler_diagnostic = test_config.MockHandler()
        cls.mock_handler_filtered_diagnostic = test_config.MockHandler()

        # Add filter to the filtered diagnostics handler

        cls.mock_handler_filtered_diagnostic.addFilter(FilterStatusCode())
        cls.logger_default = create_logger("testloggerdefault", cls.mock_handler_default)
        cls.logger_diagnostic = create_logger("testloggerdiagnostic", cls.mock_handler_diagnostic)
        cls.logger_filtered_diagnostic = create_logger("testloggerfiltereddiagnostic",
                                                       cls.mock_handler_filtered_diagnostic)
        cls.client_default = cosmos_client.CosmosClient(cls.host, cls.masterKey,
                                                        consistency_level="Session",
                                                        connection_policy=cls.connectionPolicy,
                                                        logger=cls.logger_default)
        cls.client_diagnostic = cosmos_client.CosmosClient(cls.host, cls.masterKey,
                                                           consistency_level="Session",
                                                           connection_policy=cls.connectionPolicy,
                                                           logger=cls.logger_diagnostic,
                                                           enable_diagnostics_logging=True)
        cls.client_filtered_diagnostic = cosmos_client.CosmosClient(cls.host, cls.masterKey,
                                                                    consistency_level="Session",
                                                                    connection_policy=cls.connectionPolicy,
                                                                    logger=cls.logger_filtered_diagnostic,
                                                                    enable_diagnostics_logging=True)

        # Create a root logger with a mock handler and a filter for status codes above 400
        cls.root_mock_handler = test_config.MockHandler()
        cls.root_mock_handler.addFilter(FilterStatusCode())
        cls.root_logger = create_logger("rootLogger", cls.root_mock_handler)

        # Create child loggers
        cls.root_logger_child = logging.getLogger("rootLogger.child")
        cls.root_logger_grandchild = logging.getLogger("rootLogger.child.grandchild")

        # Use the grandchild logger for the Cosmos client
        cls.client_grandchild_logger = cosmos_client.CosmosClient(
            cls.host,
            cls.masterKey,
            consistency_level="Session",
            connection_policy=cls.connectionPolicy,
            logger=cls.root_logger_grandchild,
            enable_diagnostics_logging=True
        )

    def test_default_http_logging_policy(self):
        # Test if we can log into from creating a database
        database_id = "database_test-" + str(uuid.uuid4())
        self.client_default.create_database(id=database_id)
        assert all(m.levelname == 'INFO' for m in self.mock_handler_default.messages)
        messages_request = self.mock_handler_default.messages[0].message.split("\n")
        messages_response = self.mock_handler_default.messages[1].message.split("\n")
        assert messages_request[1] == "Request method: 'GET'"
        assert 'Request headers:' in messages_request[2]
        assert messages_request[-1] == 'No body was attached to the request'
        assert messages_response[0] == 'Response status: 200'
        assert 'Response headers:' in messages_response[1]

        self.mock_handler_default.reset()

        # delete database
        self.client_default.delete_database(database_id)

    def test_cosmos_http_logging_policy(self):
        # Test if we can log info from reading a database
        database_id = "database_test-" + str(uuid.uuid4())
        self.client_diagnostic.create_database(id=database_id)
        assert all(m.levelname == 'INFO' for m in self.mock_handler_diagnostic.messages)
        # Check that we made a databaseaccount read request only once and that we only logged it once
        messages_request = self.mock_handler_diagnostic.messages[0]
        messages_response = self.mock_handler_diagnostic.messages[1]
        elapsed_time = messages_response.duration
        assert "databaseaccount" == messages_request.resource_type
        assert messages_request.verb == "GET"
        assert 200 == messages_response.status_code
        assert "Read" == messages_request.operation_type
        assert elapsed_time is not None
        assert "Response headers" in messages_response.message
        # Verify we only have a total of 6 logged messages: 2 from databaseaccount read, 2 from health check, and 2 from create database
        assert len(self.mock_handler_diagnostic.messages) == 6
        # Test if we can log into from creating a database
        # The request to create database should follow the health check request immediately
        messages_request = self.mock_handler_diagnostic.messages[4]
        messages_response = self.mock_handler_diagnostic.messages[5]
        elapsed_time = messages_response.duration
        assert "dbs" == messages_request.resource_type
        assert messages_request.verb == "POST"
        assert 201 == messages_response.status_code
        assert messages_request.operation_type == "Create"
        assert elapsed_time is not None
        assert "Response headers" in messages_response.message

        self.mock_handler_diagnostic.reset()
        # now test in case of an error
        try:
            self.client_diagnostic.create_database(id=database_id)
        except:
            pass
        assert all(m.levelname == 'INFO' for m in self.mock_handler_diagnostic.messages)
        messages_request = self.mock_handler_diagnostic.messages[0]
        messages_response = self.mock_handler_diagnostic.messages[1]
        elapsed_time = messages_response.duration
        assert "dbs" == messages_request.resource_type
        assert messages_request.operation_type == "Create"
        assert 'Request headers:' in messages_request.msg
        assert 'A body is sent with the request' in messages_request.msg
        assert messages_response.status_code == 409
        assert elapsed_time is not None
        assert "Response headers" in messages_response.msg

        # delete database
        self.client_diagnostic.delete_database(database_id)

        self.mock_handler_diagnostic.reset()

    def test_filtered_diagnostics_logging_policy(self):
        # Test if we can log errors with the filtered diagnostics logger
        database_id = "database_test_" + str(uuid.uuid4())
        container_id = "diagnostics_container_test_" + str(uuid.uuid4())
        self.client_filtered_diagnostic.create_database(id=database_id)
        database = self.client_filtered_diagnostic.get_database_client(database_id)
        database.create_container(id=container_id, partition_key=PartitionKey(path="/pk"))

        # Try to read an item that doesn't exist
        try:
            container = database.get_container_client(container_id)
            container.read_item(item="nonexistent_item", partition_key="nonexistent_pk")
        except:
            pass

        # Assert that the filtered logger only captured the error
        request_log = self.mock_handler_filtered_diagnostic.messages[0]
        response_log = self.mock_handler_filtered_diagnostic.messages[1]
        assert response_log.status_code == 404
        assert request_log.resource_type == "docs"
        assert request_log.operation_type == "Read"
        assert len(self.mock_handler_filtered_diagnostic.messages) == 2

        self.mock_handler_filtered_diagnostic.reset()

        # Try to create an item with an invalid partition key
        try:
            container.create_item(body={"FakeProperty": "item1", "NotPk": None})
        except:
            pass

        # Assert that the filtered logger captured the error
        request_log = self.mock_handler_filtered_diagnostic.messages[0]
        response_log = self.mock_handler_filtered_diagnostic.messages[1]
        assert response_log.status_code == 400
        assert request_log.resource_type == "docs"
        assert request_log.operation_type == "Create"
        assert len(self.mock_handler_filtered_diagnostic.messages) == 2

        # Clean up
        self.client_filtered_diagnostic.delete_database(database_id)
        self.mock_handler_filtered_diagnostic.reset()

    def test_client_settings(self):
        # Test data
        all_locations = [L1, L2]
        client_excluded_locations = [L1]
        multiple_write_locations = True

        # Client setup
        mock_handler = test_config.MockHandler()
        logger = create_logger("test_logger_client_settings", mock_handler)

        custom_transport = FaultInjectionTransport()
        is_get_account_predicate: Callable[[HttpRequest], bool] = lambda \
                r: FaultInjectionTransport.predicate_is_database_account_call(r)
        emulator_as_multi_write_region_account_transformation = \
            lambda r, inner: FaultInjectionTransport.transform_topology_mwr(
                first_region_name=L1,
                second_region_name=L2,
                inner=inner,
                first_region_url=L1_URL,
                second_region_url=L2_URL,
            )
        custom_transport.add_response_transformation(
            is_get_account_predicate,
            emulator_as_multi_write_region_account_transformation)

        initialized_objects = TestFaultInjectionTransport.setup_method_with_custom_transport(
            custom_transport,
            default_endpoint=CONFIG.host,
            key=CONFIG.masterKey,
            database_id=CONFIG.TEST_DATABASE_ID,
            container_id=CONFIG.TEST_SINGLE_PARTITION_CONTAINER_ID,
            preferred_locations=all_locations,
            excluded_locations=client_excluded_locations,
            multiple_write_locations=multiple_write_locations,
            custom_logger=logger
        )
        mock_handler.reset()

        # create an item
        id_value: str = str(uuid.uuid4())
        document_definition = {'id': id_value, 'pk': id_value}
        container: ContainerProxy = initialized_objects["col"]
        container.create_item(body=document_definition)

        # Verify endpoint locations
        messages_split = mock_handler.messages[1].message.split("\n")
        for message in messages_split:
            if "Preferred Regions:" in message:
                locations = get_locations_list(message)
                assert all_locations == locations
            elif "Excluded Regions:" in message:
                locations = get_locations_list(message)
                assert client_excluded_locations == locations
            elif "Account Read Regions:" in message:
                locations = get_locations_list(message)
                assert all_locations == locations
            elif "Account Write Regions:" in message:
                locations = get_locations_list(message)
                assert all_locations == locations

    def test_activity_id_logging_policy(self):
        # Create a mock handler and logger for the new client
        self.mock_handler_activity_id = test_config.MockHandler()
        self.logger_activity_id = create_logger("test_logger_activity_id", self.mock_handler_activity_id)

        # Create a new client with the logger and enable diagnostics logging
        self.client_activity_id = cosmos_client.CosmosClient(
            self.host,
            self.masterKey,
            consistency_level="Session",
            connection_policy=self.connectionPolicy,
            logger=self.logger_activity_id,
            enable_diagnostics_logging=True
        )

        # Generate a custom activity ID
        custom_activity_id = str(uuid.uuid4())

        # Create a database and container for the test
        database_id = "database_test_activity_id_" + str(uuid.uuid4())
        container_id = "container_test_activity_id_" + str(uuid.uuid4())
        database = self.client_activity_id.create_database(id=database_id)
        container = database.create_container(id=container_id, partition_key=PartitionKey(path="/pk"))
        # Reset the mock handler to clear previous messages
        self.mock_handler_activity_id.reset()

        # Upsert an item and verify the request and response activity IDs match
        item_id = str(uuid.uuid4())
        item_body = {"id": item_id, "pk": item_id}
        container.upsert_item(body=item_body)

        # Verify that the request activity ID matches the response activity ID
        # Having the Request Activity confirms we generated one from SDK
        # Having it match the response means it was passed successfully
        log_record_request = self.mock_handler_activity_id.messages[0]
        log_record_response = self.mock_handler_activity_id.messages[1]
        assert log_record_request.activity_id == log_record_response.activity_id

        # Upsert another item with the custom activity ID in the initial headers
        headers = {"x-ms-activity-id": custom_activity_id}
        item_id_2 = str(uuid.uuid4())
        item_body_2 = {"id": item_id_2, "pk": item_id_2}
        container.upsert_item(body=item_body_2, initial_headers=headers)

        # Verify that the custom activity ID does not match the request activity ID from the log record
        # Users should not be able to pass in their own activity ID.
        log_record_request_2 = self.mock_handler_activity_id.messages[2]
        assert log_record_request_2.activity_id != custom_activity_id

        # Clean up by deleting the database
        self.client_activity_id.delete_database(database_id)
        self.mock_handler_activity_id.reset()

    def test_logging_exceptions_with_no_response(self):
        # Create a mock handler and logger for capturing logs
        mock_handler = test_config.MockHandler()
        logger = create_logger("test_logger_fault_injection", mock_handler)

        # Set up FaultInjectionTransport to inject a 502 error
        id_value = str(uuid.uuid4())
        document_definition = {'id': id_value,
                               'pk': id_value,
                               'name': 'sample document',
                               'key': 'value'}
        custom_transport = FaultInjectionTransport()
        predicate: Callable[[HttpRequest], bool] = lambda r: FaultInjectionTransport.predicate_req_for_document_with_id(
            r, id_value)
        custom_transport.add_fault(predicate, lambda r: FaultInjectionTransport.error_after_delay(
            1000,
            CosmosHttpResponseError(
                status_code=502,
                message="Some random reverse proxy error.")))


        # Initialize the client with the custom transport and logger
        initialized_objects = TestFaultInjectionTransport.setup_method_with_custom_transport(
            custom_transport,
            default_endpoint=CONFIG.host,
            key=CONFIG.masterKey,
            database_id=CONFIG.TEST_DATABASE_ID,
            container_id=CONFIG.TEST_SINGLE_PARTITION_CONTAINER_ID,
            preferred_locations=[L1, L2],
            excluded_locations=[],
            multiple_write_locations=True,
            custom_logger=logger
        )
        mock_handler.reset()

        # Attempt to create an item, which should trigger the injected 502 error
        container: ContainerProxy = initialized_objects["col"]
        try:
            container.create_item(body=document_definition)
            pytest.fail("Expected exception not thrown")
        except CosmosHttpResponseError as cosmosError:
            # Verify that the logger captured the 502 error and was called from on_exception
            assert any(m.status_code == 502 and "on_exception" in m.funcName for m in mock_handler.messages)

    def test_hierarchical_logger_with_filter(self):
        # Reset the mock handler before the test
        self.root_mock_handler.reset()

        # Attempt to read a nonexistent item
        database_id = "database_test_hierarchical_logger_" + str(uuid.uuid4())
        container_id = "container_test_hierarchical_logger_" + str(uuid.uuid4())
        database = self.client_grandchild_logger.create_database(id=database_id)
        container = database.create_container(id=container_id, partition_key=PartitionKey(path="/pk"))

        try:
            container.read_item(item="nonexistent_item", partition_key="nonexistent_pk")
        except:
            pass

        # Verify that the error was logged by the root logger's mock handler
        assert len(self.root_mock_handler.messages) == 2
        log_record = self.root_mock_handler.messages[0]
        assert hasattr(log_record, "status_code")
        assert log_record.status_code == 404
        assert log_record.name == "rootLogger.child.grandchild"
        assert not bool(self.root_logger_grandchild.filters)
        assert not bool(self.root_logger_child.filters)
        assert bool(self.root_mock_handler.filters)

        # Clean up
        self.client_grandchild_logger.delete_database(database_id)
        self.root_mock_handler.reset()

if __name__ == "__main__":
    unittest.main()



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
import azure.cosmos.partition_key as partition_key
import test_config

from azure.cosmos._change_feed.feed_range_internal import FeedRangeInternalEpk
from azure.cosmos._routing.routing_range import Range

@pytest.fixture(scope="class")
def setup():
    if (TestFeedRange.masterKey == '[YOUR_KEY_HERE]' or
            TestFeedRange.host == '[YOUR_ENDPOINT_HERE]'):
        raise Exception(
            "You must specify your Azure Cosmos account values for "
            "'masterKey' and 'host' at the top of this class to run the "
            "tests.")
    test_client = cosmos_client.CosmosClient(TestFeedRange.host, test_config.TestConfig.masterKey),
    created_db = test_client[0].get_database_client(TestFeedRange.TEST_DATABASE_ID)
    return {
        "created_db": created_db,
        "created_collection": created_db.get_container_client(TestFeedRange.TEST_CONTAINER_ID)
    }

test_subset_ranges = [(Range("", "FF", True, False),
                       Range("3F", "7F", True, False),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("", "FF", True, False),
                       False),
                      (Range("3F", "7F", True, False),
                       Range("", "5F", True, False),
                       False),
                      (Range("3F", "7F", True, True),
                       Range("3F", "7F", True, True),
                       True),
                      (Range("3F", "7F", False, True),
                       Range("3F", "7F", True, True),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("3F", "7F", True, True),
                       False),
                      (Range("3F", "7F", True, False),
                       Range("", "2F", True, False),
                       False),
                      (Range("3F", "3F", True, True),
                       Range("3F", "3F", True, True),
                       True),
                      (Range("3F", "3F", True, True),
                       Range("4F", "4F", True, True),
                       False)
                      ]


test_overlaps_ranges = [(Range("", "FF", True, False),
                       Range("3F", "7F", True, False),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("", "FF", True, False),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("", "5F", True, False),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("3F", "7F", True, False),
                       True),
                      (Range("3F", "7F", True, False),
                       Range("", "2F", True, False),
                       False),
                      (Range("3F", "7F", True, False),
                       Range("6F", "FF", True, False),
                       True),
                      (Range("AA", "BB", True, False),
                       Range("CC", "FF", True, False),
                       False)]

@pytest.mark.cosmosEmulator
@pytest.mark.unittest
@pytest.mark.usefixtures("setup")
class TestFeedRange:
    """Tests to verify methods for operations on feed ranges
    """

    host = test_config.TestConfig.host
    masterKey = test_config.TestConfig.masterKey
    TEST_DATABASE_ID = test_config.TestConfig.TEST_DATABASE_ID
    TEST_CONTAINER_ID = test_config.TestConfig.TEST_MULTI_PARTITION_CONTAINER_ID


    def test_partition_key_to_feed_range(self, setup):
        created_container = setup["created_db"].create_container(
            id='container_' + str(uuid.uuid4()),
            partition_key=partition_key.PartitionKey(path="/id")
        )
        feed_range = created_container.feed_range_from_partition_key("1")
        feed_range_epk = FeedRangeInternalEpk.from_json(feed_range)
        assert feed_range_epk.get_normalized_range() == Range("3C80B1B7310BB39F29CC4EA05BDD461E",
                        "3c80b1b7310bb39f29cc4ea05bdd461f", True, False)
        setup["created_db"].delete_container(created_container)

    @pytest.mark.parametrize("parent_feed_range, child_feed_range, is_subset", test_subset_ranges)
    def test_feed_range_is_subset(self, setup, parent_feed_range, child_feed_range, is_subset):
        epk_parent_feed_range = FeedRangeInternalEpk(parent_feed_range).to_dict()
        epk_child_feed_range = FeedRangeInternalEpk(child_feed_range).to_dict()
        assert setup["created_collection"].is_feed_range_subset(epk_parent_feed_range, epk_child_feed_range) == is_subset

    def test_feed_range_is_subset_from_pk(self, setup):
        epk_parent_feed_range = FeedRangeInternalEpk(
            Range("", "FF", True, False)).to_dict()
        epk_child_feed_range = setup["created_collection"].feed_range_from_partition_key("1")
        assert setup["created_collection"].is_feed_range_subset(epk_parent_feed_range, epk_child_feed_range)

    @pytest.mark.parametrize("range1, range2, overlaps", test_overlaps_ranges)
    def test_overlaps(self, setup, range1, range2, overlaps):
        assert Range.overlaps(range1, range2) == overlaps

if __name__ == '__main__':
    unittest.main()



# The MIT License (MIT)
# Copyright (c) 2018 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import asyncio # pylint: disable=C4763  # Used for Semaphore and gather, not for sleep
from typing import Tuple, Any, Sequence, Optional, TYPE_CHECKING, Mapping

from azure.cosmos import _base, exceptions
from azure.core.utils import CaseInsensitiveDict
from azure.cosmos._query_builder import _QueryBuilder
from azure.cosmos.partition_key import _get_partition_key_from_partition_key_definition, PartitionKeyType
from azure.cosmos import CosmosList

if TYPE_CHECKING:
    from azure.cosmos.aio._cosmos_client_connection_async import CosmosClientConnection

class ReadItemsHelperAsync:
    """Helper class for handling read many items operations.

    This implementation preserves the original order of items in the input sequence
    when returning results, regardless of how items are distributed across partitions
    or processed in chunks.
    """
    logger = logging.getLogger("azure.cosmos.ReadManyItemsHelper")

    def __init__(
            self,
            client: 'CosmosClientConnection',
            collection_link: str,
            items: Sequence[Tuple[str, PartitionKeyType]],
            options: Optional[Mapping[str, Any]],
            partition_key_definition: dict[str, Any],
            max_concurrency: int = 5,
            **kwargs: Any
    ):
        self.client = client
        self.collection_link = collection_link
        self.items = items
        self.options = dict(options) if options is not None else {}
        self.partition_key_definition = partition_key_definition
        self.kwargs = kwargs
        self.max_concurrency = max_concurrency if max_concurrency and max_concurrency > 0 else 5
        self.max_items_per_query = 1000

    async def read_items(self) -> 'CosmosList':
        """Executes the read-many operation.

        :return: A list of the retrieved items in the same order as the input.
        :rtype: ~azure.cosmos.CosmosList
        """
        if not self.items:
            return CosmosList([], response_headers=CaseInsensitiveDict())

        items_by_partition = await self._partition_items_by_range()
        if not items_by_partition:
            return CosmosList([], response_headers=CaseInsensitiveDict())

        query_chunks = self._create_query_chunks(items_by_partition)

        indexed_results, combined_headers = await self._execute_queries_concurrently(query_chunks)

        indexed_results.sort(key=lambda x: x[0])
        all_results = [item[1] for item in indexed_results]
        cosmos_list = CosmosList(all_results, response_headers=combined_headers)

        if 'response_hook' in self.kwargs:
            self.kwargs['response_hook'](combined_headers, cosmos_list)

        return cosmos_list

    async def _partition_items_by_range(self) -> dict[str, list[Tuple[int, str, "PartitionKeyType"]]]:
        # pylint: disable=protected-access
        """
        Groups items by their partition key range ID efficiently while preserving original order.

        :return: A dictionary mapping partition key range IDs to lists of tuples containing the
                    original index, item ID, and partition key value.
        :rtype: dict[str, list[tuple[int, str, PartitionKeyType]]]
        """
        collection_rid = _base.GetResourceIdOrFullNameFromLink(self.collection_link)
        partition_key = _get_partition_key_from_partition_key_definition(self.partition_key_definition)
        items_by_partition: dict[str, list[Tuple[int, str, "PartitionKeyType"]]] = {}

        items_by_pk_value: dict[Any, list[Tuple[int, str, "PartitionKeyType"]]] = {}
        for idx, (item_id, pk_value) in enumerate(self.items):
            key = tuple(pk_value) if isinstance(pk_value, list) else pk_value
            if key not in items_by_pk_value:
                items_by_pk_value[key] = []
            items_by_pk_value[key].append((idx, item_id, pk_value))

        for pk_items in items_by_pk_value.values():
            pk_value = pk_items[0][2]
            epk_range = partition_key._get_epk_range_for_partition_key(pk_value)
            overlapping_ranges = await self.client._routing_map_provider.get_overlapping_ranges(
                collection_rid, [epk_range], self.options
            )
            if overlapping_ranges:
                range_id = overlapping_ranges[0]["id"]
                if range_id not in items_by_partition:
                    items_by_partition[range_id] = []
                items_by_partition[range_id].extend(pk_items)

        return items_by_partition

    def _create_query_chunks(
            self,
            items_by_partition: dict[str, list[Tuple[int, str, "PartitionKeyType"]]]
    ) -> list[dict[str, list[Tuple[int, str, "PartitionKeyType"]]]]:

        """
        Create query chunks for concurrency control while preserving original indices.

        :param items_by_partition: A dictionary mapping partition key range IDs to lists of tuples containing the
                                original index, item ID, and partition key value.
        :type items_by_partition: dict[str, list[tuple[int, str, PartitionKeyType]]]
        :return: A list of dictionaries, each mapping a partition ID to a chunk of items.
        :rtype: list[dict[str, list[tuple[int, str, PartitionKeyType]]]]
        """
        query_chunks = []
        for partition_id, partition_items in items_by_partition.items():
            for i in range(0, len(partition_items), self.max_items_per_query):
                chunk = partition_items[i:i + self.max_items_per_query]
                query_chunks.append({partition_id: chunk})
        return query_chunks

    async def _execute_queries_concurrently(
            self,
            query_chunks: list[dict[str, list[Tuple[int, str, "PartitionKeyType"]]]],
    ) -> Tuple[list[Tuple[int, Any]], CaseInsensitiveDict]:
        """
        Execute query chunks concurrently and return aggregated results with original indices.

        :param query_chunks: A list of dictionaries, each mapping a partition ID to a chunk of items to query.
        :type query_chunks: list[dict[str, list[tuple[int, str, PartitionKeyType]]]]
        :return: A tuple containing a list of results with original indices and the combined response headers.
        :rtype: tuple[list[tuple[int, any]], CaseInsensitiveDict]
        """
        if not query_chunks:
            return [], CaseInsensitiveDict()

        semaphore = asyncio.Semaphore(self.max_concurrency)
        indexed_results = []
        total_request_charge = 0.0

        async def execute_chunk_query(partition_id, chunk_partition_items):
            async with semaphore:
                id_to_idx = {item[1]: item[0] for item in chunk_partition_items}
                items_for_query = [(item[1], item[2]) for item in chunk_partition_items]
                request_kwargs = self.kwargs.copy()

                if len(items_for_query) == 1:
                    item_id, pk_value = items_for_query[0]
                    result, headers = await self._execute_point_read(item_id, pk_value, request_kwargs)
                    chunk_results = [(id_to_idx[item_id], result)] if result else []
                else:
                    chunk_results, headers = await self._execute_query(
                        partition_id, items_for_query, id_to_idx, request_kwargs)

                request_charge = self._extract_request_charge(headers)
                return chunk_results, request_charge

        tasks = [
            asyncio.create_task(execute_chunk_query(partition_id, items))
            for chunk in query_chunks
            for partition_id, items in chunk.items()
        ]

        try:
            all_chunk_results = await asyncio.gather(*tasks)
        except Exception:
            for task in tasks:
                if not task.done():
                    task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise

        for chunk_result, ru_charge in all_chunk_results:
            indexed_results.extend(chunk_result)
            total_request_charge += ru_charge

        final_headers = CaseInsensitiveDict({'x-ms-request-charge': str(total_request_charge)})
        return indexed_results, final_headers

    def _extract_request_charge(self, headers: CaseInsensitiveDict) -> float:
        """Extract the request charge from the headers.

        :param headers: The response headers.
        :type headers: ~azure.core.utils.CaseInsensitiveDict
        :return: The request charge.
        :rtype: float
        """
        charge = headers.get('x-ms-request-charge')
        request_charge = 0.0  # Renamed from ru_charge to avoid shadowing
        if charge:
            try:
                request_charge = float(charge)
            except (ValueError, TypeError):
                self.logger.warning("Invalid request charge format: %s", charge)
        return request_charge

    async def _execute_point_read(
            self,
            item_id: str,
            pk_value: "PartitionKeyType",
            request_kwargs: dict[str, Any]
    ) -> Tuple[Optional[Any], CaseInsensitiveDict]:
        """
        Executes a point read for a single item.

        :param item_id: The ID of the item to read.
        :type item_id: str
        :param pk_value: The partition key value for the item.
        :type pk_value: PartitionKeyType
        :param request_kwargs: Additional keyword arguments for the request.
        :type request_kwargs: dict[str, any]
        :return: A tuple containing the item (or None if not found) and the response headers.
        :rtype: tuple[Optional[any], CaseInsensitiveDict]
        """
        doc_link = f"{self.collection_link}/docs/{item_id}"
        point_read_options = self.options.copy()
        point_read_options["partitionKey"] = pk_value
        captured_headers = {}

        def local_response_hook(hook_headers, _):
            captured_headers.update(hook_headers)

        request_kwargs['response_hook'] = local_response_hook
        request_kwargs.pop("containerProperties", None)

        try:
            result = await self.client.ReadItem(doc_link, point_read_options, **request_kwargs)
            return result, CaseInsensitiveDict(captured_headers)
        except exceptions.CosmosResourceNotFoundError as e:
            captured_headers.update(e.headers)
            return None, CaseInsensitiveDict(captured_headers)

    async def _execute_query(
            self,
            partition_id: str,
            items_for_query: Sequence[Tuple[str, "PartitionKeyType"]],
            id_to_idx: dict[str, int],
            request_kwargs: dict[str, Any]
    ) -> Tuple[list[Tuple[int, Any]], CaseInsensitiveDict]:
        """
        Builds and executes a query for a chunk of items.

        :param partition_id: The partition key range ID for the query.
        :type partition_id: str
        :param items_for_query: A list of tuples containing item IDs and partition key values to query.
        :type items_for_query: list[tuple[str, PartitionKeyType]]
        :param id_to_idx: A mapping from item ID to its original index in the input sequence.
        :type id_to_idx: dict[str, int]
        :param request_kwargs: Additional keyword arguments for the request.
        :type request_kwargs: dict[str, any]
        :return: A tuple containing a list of results with original indices and the response headers.
        :rtype: tuple[list[tuple[int, any]], CaseInsensitiveDict]
        """
        captured_headers = {}

        def local_response_hook(hook_headers, _):
            captured_headers.update(hook_headers)

        request_kwargs['response_hook'] = local_response_hook

        if _QueryBuilder.is_id_partition_key_query(items_for_query, self.partition_key_definition):
            query_obj = _QueryBuilder.build_id_in_query(items_for_query)
        elif _QueryBuilder.is_single_logical_partition_query(items_for_query):
            query_obj = _QueryBuilder.build_pk_and_id_in_query(items_for_query, self.partition_key_definition)
        else:
            partition_items_dict = {partition_id: items_for_query}
            query_obj = _QueryBuilder.build_parameterized_query_for_items(
                partition_items_dict, self.partition_key_definition)

        page_iterator = self.client.QueryItems(
            self.collection_link, query_obj, self.options, **request_kwargs).by_page()

        chunk_indexed_results = []
        async for page in page_iterator:
            async for item in page:
                doc_id = item.get('id')
                if doc_id in id_to_idx:
                    chunk_indexed_results.append((id_to_idx[doc_id], item))
                else:
                    self.logger.warning("Received document with unexpected ID: %s", doc_id)

        return chunk_indexed_results, CaseInsensitiveDict(captured_headers)



# The MIT License (MIT)
# Copyright (c) 2014 Microsoft Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal class for partition key range implementation in the Azure Cosmos
database service.
"""
import base64
import binascii
import json


class PartitionKeyRange(object):
    """Partition Key Range Constants"""

    MinInclusive = "minInclusive"
    MaxExclusive = "maxExclusive"
    Id = "id"
    Parents = "parents"


class Range(object):
    """description of class"""

    MinPath = "min"
    MaxPath = "max"
    IsMinInclusivePath = "isMinInclusive"
    IsMaxInclusivePath = "isMaxInclusive"

    def __init__(self, range_min, range_max, isMinInclusive, isMaxInclusive):
        if range_min is None:
            raise ValueError("min is missing")
        if range_max is None:
            raise ValueError("max is missing")

        self.min = range_min.upper()
        self.max = range_max.upper()
        self.isMinInclusive = isMinInclusive
        self.isMaxInclusive = isMaxInclusive

    def contains(self, value):
        minToValueRelation = self.min > value
        maxToValueRelation = self.max > value
        return (
            (self.isMinInclusive and minToValueRelation <= 0) or (not self.isMinInclusive and minToValueRelation < 0)
        ) and (
            (self.isMaxInclusive and maxToValueRelation >= 0) or (not self.isMaxInclusive and maxToValueRelation > 0)
        )

    @classmethod
    def get_full_range(cls):
        """Gets a Range object that covers the entire possible range of partition key values.

        :return: A Range object that covers the entire possible range of partition key values.
        :rtype: ~azure.cosmos._routing.routing_range.Range
        """
        return cls(range_min="", range_max="FF", isMinInclusive=True, isMaxInclusive=False)

    @classmethod
    def PartitionKeyRangeToRange(cls, partition_key_range):
        self = cls(
            partition_key_range[PartitionKeyRange.MinInclusive].upper(),
            partition_key_range[PartitionKeyRange.MaxExclusive].upper(),
            True,
            False,
        )
        return self

    @classmethod
    def ParseFromDict(cls, range_as_dict):
        self = cls(
            range_as_dict[Range.MinPath].upper(),
            range_as_dict[Range.MaxPath].upper(),
            range_as_dict[Range.IsMinInclusivePath],
            range_as_dict[Range.IsMaxInclusivePath],
        )
        return self

    def to_dict(self):
        return {
            self.MinPath: self.min,
            self.MaxPath: self.max,
            self.IsMinInclusivePath: self.isMinInclusive,
            self.IsMaxInclusivePath: self.isMaxInclusive
        }

    def to_normalized_range(self):
        if self.isMinInclusive and not self.isMaxInclusive:
            return self

        normalized_min = self.min
        normalized_max = self.max

        if not self.isMinInclusive:
            normalized_min = self.add_to_effective_partition_key(self.min, -1)

        if self.isMaxInclusive:
            normalized_max = self.add_to_effective_partition_key(self.max, 1)

        return Range(normalized_min, normalized_max, True, False)

    def add_to_effective_partition_key(self, effective_partition_key: str, value: int):
        if value not in (-1, 1):
            raise ValueError("Invalid value - only 1 or -1 is allowed")

        byte_array = self.hex_binary_to_byte_array(effective_partition_key)
        if value == 1:
            for i in range(len(byte_array) -1, -1, -1):
                if byte_array[i] < 255:
                    byte_array[i] += 1
                    break
                byte_array[i] = 0
        else:
            for i in range(len(byte_array) - 1, -1, -1):
                if byte_array[i] != 0:
                    byte_array[i] -= 1
                    break
                byte_array[i] = 255

        return binascii.hexlify(byte_array).decode().upper()

    def hex_binary_to_byte_array(self, hex_binary_string: str):
        if hex_binary_string is None:
            raise ValueError("hex_binary_string is missing")
        if len(hex_binary_string) % 2 != 0:
            raise ValueError("hex_binary_string must not have an odd number of characters")

        return bytearray.fromhex(hex_binary_string)

    @classmethod
    def from_base64_encoded_json_string(cls, data: str):
        try:
            feed_range_json_string = base64.b64decode(data, validate=True).decode('utf-8')
            feed_range_json = json.loads(feed_range_json_string)
            return cls.ParseFromDict(feed_range_json)
        except Exception as exc:
            raise ValueError(f"Invalid feed_range json string {data}") from exc

    def to_base64_encoded_string(self):
        data_json = json.dumps(self.to_dict())
        json_bytes = data_json.encode('utf-8')
        # Encode the bytes to a Base64 string
        base64_bytes = base64.b64encode(json_bytes)
        # Convert the Base64 bytes to a string
        return base64_bytes.decode('utf-8')

    def isSingleValue(self):
        return self.isMinInclusive and self.isMaxInclusive and self.min == self.max

    def isEmpty(self):
        return (not (self.isMinInclusive and self.isMaxInclusive)) and self.min == self.max

    def __hash__(self):
        return hash((self.min, self.max, self.isMinInclusive, self.isMaxInclusive))

    def __str__(self):

        return (
            ("[" if self.isMinInclusive else "(")
            + str(self.min)
            + ","
            + str(self.max)
            + ("]" if self.isMaxInclusive else ")")
        )

    def __eq__(self, other):
        return (
            (self.min == other.min)
            and (self.max == other.max)
            and (self.isMinInclusive == other.isMinInclusive)
            and (self.isMaxInclusive == other.isMaxInclusive)
        )

    @staticmethod
    def _compare_helper(a: str, b: str):
        # python 3 compatible
        return (a > b) - (a < b)

    @staticmethod
    def overlaps(range1, range2):
        if range1 is None or range2 is None:
            return False
        if range1.isEmpty() or range2.isEmpty():
            return False

        cmp1 = Range._compare_helper(range1.min, range2.max)
        cmp2 = Range._compare_helper(range2.min, range1.max)

        if cmp1 <= 0 and cmp2 <= 0:
            if (cmp1 == 0 and not (range1.isMinInclusive and range2.isMaxInclusive)) or (
                cmp2 == 0 and not (range2.isMinInclusive and range1.isMaxInclusive)
            ):
                return False
            return True
        return False

    def can_merge(self, other: 'Range') -> bool:
        if self.isSingleValue() and other.isSingleValue():
            return self.min == other.min
        # if share the same boundary, they can merge
        overlap_boundary1 = self.max == other.min and self.isMaxInclusive or other.isMinInclusive
        overlap_boundary2 = other.max == self.min and other.isMaxInclusive or self.isMinInclusive
        if overlap_boundary1 or overlap_boundary2:
            return True
        return self.overlaps(self, other)

    def merge(self, other: 'Range') -> 'Range':
        if not self.can_merge(other):
            raise ValueError("Ranges do not overlap")
        min_val = self.min if self.min < other.min else other.min
        max_val = self.max if self.max > other.max else other.max
        is_min_inclusive = self.isMinInclusive if self.min < other.min else other.isMinInclusive
        is_max_inclusive = self.isMaxInclusive if self.max > other.max else other.isMaxInclusive
        return Range(min_val, max_val, is_min_inclusive, is_max_inclusive)

    def is_subset(self, parent_range: 'Range') -> bool:
        normalized_parent_range = parent_range.to_normalized_range()
        normalized_child_range = self.to_normalized_range()
        return (normalized_parent_range.min <= normalized_child_range.min and
                normalized_parent_range.max >= normalized_child_range.max)

class PartitionKeyRangeWrapper(object):
    """Internal class for a representation of a unique partition for an account
    """

    def __init__(self, partition_key_range: Range, collection_rid: str) -> None:
        self.partition_key_range = partition_key_range
        self.collection_rid = collection_rid


    def __str__(self) -> str:
        return (
            f"PartitionKeyRangeWrapper("
            f"partition_key_range={self.partition_key_range}, "
            f"collection_rid={self.collection_rid}, "
        )

    def __eq__(self, other):
        if not isinstance(other, PartitionKeyRangeWrapper):
            return False
        return self.partition_key_range == other.partition_key_range and self.collection_rid == other.collection_rid

    def __hash__(self):
        return hash((self.partition_key_range, self.collection_rid))



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.

import unittest
import uuid

import pytest

import azure.cosmos.cosmos_client as cosmos_client
from azure.cosmos import http_constants, exceptions, PartitionKey
from test_config import TestConfig


# This test class serves to test user-configurable options and verify they are
# properly set and saved into the different object instances that use these
# user-configurable settings.


def get_test_item():
    item = {
        'id': 'Async_' + str(uuid.uuid4()),
        'test_object': True,
        'lastName': 'Smith'
    }
    return item


@pytest.mark.cosmosLong
class TestUserConfigs(unittest.TestCase):

    def test_invalid_connection_retry_configuration(self):
        try:
            cosmos_client.CosmosClient(url=TestConfig.host, credential=TestConfig.masterKey,
                                       consistency_level="Session", connection_retry_policy="Invalid Policy")
        except TypeError as e:
            self.assertTrue(str(e).startswith('Unsupported retry policy'))

    def test_enable_endpoint_discovery(self):
        client_false = cosmos_client.CosmosClient(url=TestConfig.host, credential=TestConfig.masterKey,
                                                  consistency_level="Session", enable_endpoint_discovery=False)
        client_default = cosmos_client.CosmosClient(url=TestConfig.host, credential=TestConfig.masterKey,
                                                    consistency_level="Session")
        client_true = cosmos_client.CosmosClient(url=TestConfig.host, credential=TestConfig.masterKey,
                                                 consistency_level="Session", enable_endpoint_discovery=True)

        self.assertFalse(client_false.client_connection.connection_policy.EnableEndpointDiscovery)
        self.assertTrue(client_default.client_connection.connection_policy.EnableEndpointDiscovery)
        self.assertTrue(client_true.client_connection.connection_policy.EnableEndpointDiscovery)

    def test_authentication_error(self):
        try:
            cosmos_client.CosmosClient(url=TestConfig.host, credential="wrong_key")
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, http_constants.StatusCodes.UNAUTHORIZED)

    def test_default_account_consistency(self):
        database_id = "PythonSDKUserConfigTesters-" + str(uuid.uuid4())
        container_id = "PythonSDKTestContainer-" + str(uuid.uuid4())
        client = cosmos_client.CosmosClient(url=TestConfig.host, credential=TestConfig.masterKey)
        database_account = client.get_database_account()
        account_consistency_level = database_account.ConsistencyPolicy["defaultConsistencyLevel"]
        self.assertEqual(account_consistency_level, "Session")

        # Testing the session token logic works without user passing in Session explicitly
        database = client.create_database(database_id)
        container = database.create_container(id=container_id, partition_key=PartitionKey(path="/id"))
        create_response = container.create_item(body=get_test_item())
        session_token = create_response.get_response_headers()[http_constants.CookieHeaders.SessionToken]
        item2 = get_test_item()
        create_response = container.create_item(body=item2)
        session_token2 = create_response.get_response_headers()[http_constants.CookieHeaders.SessionToken]

        # Check Session token is being updated to reflect new item created
        self.assertNotEqual(session_token, session_token2)

        read_response = container.read_item(item=item2.get("id"), partition_key=item2.get("id"))
        read_session_token = read_response.get_response_headers()[http_constants.CookieHeaders.SessionToken]

        # Check Session token remains the same for read operation as with previous create item operation
        self.assertEqual(session_token2, read_session_token)
        client.delete_database(database_id)

        # Now testing a user-defined consistency level as opposed to using the account one
        custom_level = "Eventual"
        eventual_consistency_client = cosmos_client.CosmosClient(url=TestConfig.host,
                                                                 credential=TestConfig.masterKey,
                                                                 consistency_level=custom_level)
        database_account = eventual_consistency_client.get_database_account()
        account_consistency_level = database_account.ConsistencyPolicy["defaultConsistencyLevel"]
        # Here they're not equal, since the headers being used make the client use a different level of consistency
        self.assertNotEqual(
            eventual_consistency_client
            .client_connection.default_headers[http_constants.HttpHeaders.ConsistencyLevel],
            account_consistency_level)

        # Test for failure when trying to set consistency to higher level than account level
        custom_level = "Strong"
        strong_consistency_client = cosmos_client.CosmosClient(url=TestConfig.host,
                                                               credential=TestConfig.masterKey,
                                                               consistency_level=custom_level)
        try:
            strong_consistency_client.create_database(database_id)
        except exceptions.CosmosHttpResponseError as e:
            self.assertEqual(e.status_code, http_constants.StatusCodes.BAD_REQUEST)


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import asyncio
from azure.cosmos import exceptions, PartitionKey
from azure.cosmos.aio import CosmosClient

import json
import os


async def examples_async():
    # All interaction with Cosmos DB starts with an instance of the CosmosClient
    # In order to use the asynchronous client, we need to use async/await keywords,
    # which can only be used within async methods like examples_async() here

    # Since this is an asynchronous client, in order to properly use it you also have to warm it up and close it down.
    # We recommend using the `async with` keywords like below to start your clients - these keywords
    # create a context manager that automatically warms up, initializes, and cleans up the client, so you don't have to.

    # [START create_client]
    url = os.environ["ACCOUNT_URI"]
    key = os.environ["ACCOUNT_KEY"]
    async with CosmosClient(url, key) as client:
        # [END create_client]

        # Create a database in the account using the CosmosClient,
        # specifying that the operation shouldn't throw an exception
        # if a database with the given ID already exists.
        # [START create_database]
        database_name = "testDatabase"
        try:
            database = await client.create_database(id=database_name)
        except exceptions.CosmosResourceExistsError:
            database = client.get_database_client(database=database_name)
        # [END create_database]

        # Create a container, handling the exception if a container with the
        # same ID (name) already exists in the database.
        # [START create_container]
        container_name = "products"
        try:
            container = await database.create_container(
                id=container_name, partition_key=PartitionKey(path="/productName")
            )
        except exceptions.CosmosResourceExistsError:
            container = database.get_container_client(container_name)
        # [END create_container]

        # Create a container with custom settings. This example
        # creates a container with a custom partition key.
        # [START create_container_with_settings]
        customer_container_name = "customers"
        try:
            customer_container = await database.create_container(
                id=customer_container_name,
                partition_key=PartitionKey(path="/city"),
                default_ttl=200,
            )
        except exceptions.CosmosResourceExistsError:
            customer_container = database.get_container_client(customer_container_name)
        # [END create_container_with_settings]

        # Retrieve a container by walking down the resource hierarchy
        # (client->database->container), handling the exception generated
        # if no container with the specified ID was found in the database.
        # [START get_container]
        database = client.get_database_client(database_name)
        container_client = database.get_container_client(container_name)
        # [END get_container]

        # [START list_containers]
        database = client.get_database_client(database_name)
        async for container_c in database.list_containers():
            print("Container ID: {}".format(container_c['id']))
        # [END list_containers]

        # Insert new items by defining a dict and calling Container.upsert_item
        # [START upsert_items]
        container = database.get_container_client(container_name)
        for i in range(1, 10):
            await container.upsert_item(
                dict(id="item{}".format(i), productName="Widget", productModel="Model {}".format(i))
            )
        # [END upsert_items]

        # Modify an existing item in the container
        # [START update_item]
        item = await container.read_item("item2", partition_key="Widget")
        item["productModel"] = "DISCONTINUED"
        updated_item = await container.upsert_item(item)
        # [END update_item]

        # Query the items in a container using SQL-like syntax. This example
        # gets all items whose product model hasn't been discontinued.
        # The asynchronous client returns asynchronous iterators for its query methods;
        # as such, we iterate over it by using an async for loop
        # [START query_items]
        async for queried_item in container.query_items(
                query='SELECT * FROM products p WHERE p.productModel <> "DISCONTINUED"'
        ):
            print(json.dumps(queried_item, indent=True))
        # [END query_items]

        # Parameterized queries are also supported. This example
        # gets all items whose product model has been discontinued.
        # [START query_items_param]
        discontinued_items = container.query_items(
            query='SELECT * FROM products p WHERE p.productModel = @model AND p.productName="Widget"',
            parameters=[dict(name="@model", value="DISCONTINUED")],
        )
        async for discontinued_item in discontinued_items:
            print(json.dumps(discontinued_item, indent=True))
        # [END query_items_param]

        # [START priority_level option]
        # Priority-based execution is a capability which allows users to specify priority
        # for the request sent to Azure Cosmos DB. Based on the priority specified by the user,
        # if there are more requests than the configured RU/s in a second,
        # then Azure Cosmos DB will throttle low priority requests to allow high priority requests to execute.
        # Can be used for Read, Write, and Query operations. This is specified with the `priority` keyword.
        # the value can either be low or high.
        
        async for queried_item in container.query_items(
                query='SELECT * FROM products p WHERE p.productModel <> "DISCONTINUED"', priority="High"
        ):
            print(json.dumps(queried_item, indent=True))
        # [END priority_level option]

        # [START client_level_priority]
        # Priority can also be set at the client level, which will apply to all requests made by that client.
        # This is useful when you want all operations from a particular client to have the same priority.
        # The client-level priority is set during client initialization with the `priority` parameter.

        # Create a client with Low priority for all requests
        async with CosmosClient(url, key, priority="Low") as low_priority_client:
            low_priority_database = low_priority_client.get_database_client(database_name)
            low_priority_container = low_priority_database.get_container_client(container_name)

            # Add some items to query
            for i in range(1, 4):
                await low_priority_container.upsert_item(
                    dict(id="low_priority_item{}".format(i), productName="Widget", productModel="Model {}".format(i))
                )

            # All requests from this client will have Low priority by default
            async for queried_item in low_priority_container.query_items(
                query='SELECT * FROM products p WHERE p.productName = "Widget"'
            ):
                print(json.dumps(queried_item, indent=True))

        # [END client_level_priority]

        # [START request_priority_precedence]
        # Request-level priority takes precedence over client-level priority.
        # This allows you to override the default priority for specific operations.

        # Create a client with Low priority
        async with CosmosClient(url, key, priority="Low") as client_with_default_priority:
            database_with_priority = client_with_default_priority.get_database_client(database_name)
            container_with_priority = database_with_priority.get_container_client(container_name)

            # Add items with different priority levels to the container
            await container_with_priority.upsert_item(
                dict(id="urgent_item1", productName="Widget", priority="Low", productModel="Low Priority Model")
            )
            await container_with_priority.upsert_item(
                dict(id="normal_item1", productName="Widget", priority="High", productModel="High Priority Model")
            )

            # This query will use High priority, overriding the client's Low priority setting
            async for important_item in container_with_priority.query_items(
                query='SELECT * FROM products p WHERE p.priority = "High"',
                priority="High"  # Request-level priority overrides client-level priority
            ):
                print(json.dumps(important_item, indent=True))

            # This query will use the client's default Low priority
            async for normal_item in container_with_priority.query_items(
                query='SELECT * FROM products p WHERE p.priority = "Low"'
            ):
                print(json.dumps(normal_item, indent=True))

        # [END request_priority_precedence]

        # Delete items from the container.
        # The Cosmos DB SQL API does not support 'DELETE' queries,
        # so deletes must be done with the delete_item method
        # on the container.
        # [START delete_items]
        async for queried_item in container.query_items(
                query='SELECT * FROM products p WHERE p.productModel = "DISCONTINUED" AND p.productName="Widget"'
        ):
            await container.delete_item(queried_item, partition_key="Widget")
        # [END delete_items]

        # Retrieve the properties of a database
        # [START get_database_properties]
        read_properties = await database.read()
        print(json.dumps(read_properties, indent=True))
        # [END get_database_properties]

        # Retrieve the properties of a container
        # [START get_container_properties]
        # Get properties will return a cache of two container properties: RID and the Partition Key Definition (This will not consume RUs)
        properties = await container._get_properties()

        # Print _rid and partitionKey
        print("Resource ID: ", properties.get('_rid'))
        print("Partition Key: ", properties.get('partitionKey'))

        # Read the container to get the latests of all the Container Properties. (This will make a backend requests and will consume RUs)
        container_properties = await container.read()

        # Print each property one by one if they are currently in the container properties
        print("indexingPolicy: ", container_properties.get("indexingPolicy"))
        print("etag: ", container_properties.get('_etag'))
        print("lastModified: ", container_properties.get('lastModified'))
        print("defaultTtl: ", container_properties.get('defaultTtl'))
        print("uniqueKeyPolicy: ", container_properties.get('uniqueKeyPolicy'))
        print("conflictResolutionPolicy: ", container_properties.get('conflictResolutionPolicy'))
        print("changeFeedPolicy: ", container_properties.get('changeFeedPolicy'))
        print("geospatialConfig: ", container_properties.get('geospatialConfig'))

        # Print remaining properties if they are in the current container properties
        for key, value in container_properties.items():
            if key not in ['_rid', 'partitionKey', 'indexingPolicy', '_etag', 'lastModified', 'defaultTtl', 'uniqueKeyPolicy',
                           'conflictResolutionPolicy', 'changeFeedPolicy', 'geospatialConfig']:
                print(f"{key}: {value}")
        # [END get_container_properties]

        # Modify the properties of an existing container
        # This example sets the default time to live (TTL) for items in the
        # container to 3600 seconds (1 hour). An item in container is deleted
        # when the TTL has elapsed since it was last edited.
        # [START reset_container_properties]
        # Set the TTL on the container to 3600 seconds (one hour)
        await database.replace_container(container, partition_key=PartitionKey(path='/productName'), default_ttl=3600)

        # Display the new TTL setting for the container
        container_props = await database.get_container_client(container_name).read()
        print("New container TTL: {}".format(json.dumps(container_props['defaultTtl'])))
        # [END reset_container_properties]

        # Create a user in the database.
        # [START create_user]
        try:
            await database.create_user(dict(id="Walter Harp"))
            print("Created user Walter Harp.")
        except exceptions.CosmosResourceExistsError:
            print("A user with that ID already exists.")
        except exceptions.CosmosHttpResponseError as failure:
            print("Failed to create user. Status code:{}".format(failure.status_code))
        # [END create_user]

        # delete all items in a given partition key
        # [START delete_all_items_by_partition_key]
        container_name = "products"
        container = database.get_container_client(container_name)
        for i in range(1, 10):
            await container.upsert_item(
                dict(id="item{}".format(i), productName="Gadget", productModel="Model {}".format(i))
            )
        items = container.read_all_items()
        async for item_dict in items:
            print(json.dumps(item_dict, indent=True))
        await container.delete_all_items_by_partition_key("Gadget")
        print("All items in partition {} deleted.".format("Gadget"))
        items = container.read_all_items()
        async for item_dict in items:
            print(json.dumps(item_dict, indent=True))
        # [END delete_all_items_by_partition_key]

        # insert items in a subpartitioned container
        # [START create_container]
        location_container_name = "locations"
        try:
            container = await database.create_container(
                id=location_container_name,
                partition_key=PartitionKey(path=["/state", "/city", "/zipcode"], kind="MultiHash")
            )
        except exceptions.CosmosResourceExistsError:
            container = database.get_container_client(location_container_name)
        # [END create_container]
        # [START upsert_items]
        for i in range(1, 10):
            await container.upsert_item(
                dict(id="item{}".format(i), state="WA", city="Redmond", zipcode=98052)
            )
        # [END upsert_items]

        # Modify an existing item in the container
        # [START update_item]
        item = await container.read_item("item2", partition_key=["WA", "Redmond", 98052])
        item["state"] = "GA"
        item["city"] = "Atlanta"
        item["zipcode"] = 30363
        updated_item = await container.upsert_item(item)
        # [END update_item]

        # Query the items in a container using SQL-like syntax. This example
        # gets all items whose product model hasn't been discontinued.
        # [START query_items]
        async for queried_item in container.query_items(
                query='SELECT * FROM location l WHERE l.state = "WA"'
        ):
            print(json.dumps(queried_item, indent=True))
        # [END query_items]

         # [START delete_items]
        async for item_dict in container.query_items(
                query='SELECT * FROM location p WHERE p.state = "GA"'
        ):
            await container.delete_item(item_dict, partition_key=["GA", "Atlanta", 30363])
        # [END delete_items]

        # Query items with feed range is also supported. This example
        # gets all items within the feed range.
        # [START query_items_feed_range]
        async for feed_range in container.read_feed_ranges():
            async for queried_item in container.query_items(
                    query='SELECT * from c',
                    feed_range=feed_range):
                print(json.dumps(queried_item, indent=True))
        # [END query_items_param]

        # Get the feed ranges list from container.
        # [START read_feed_ranges]
        feed_ranges = [feed_range async for feed_range in container.read_feed_ranges()]
        # [END read_feed_ranges]

        # Get a feed range from a partition key.
        # [START feed_range_from_partition_key ]
        feed_range_from_pk = await container.feed_range_from_partition_key(["GA", "Atlanta", 30363])
        # [END feed_range_from_partition_key]

        # Figure out if a feed range is a subset of another feed range.
        # This example sees in which feed range from the container a feed range from a partition key is part of.
        # [START is_feed_range_subset]
        parent_feed_range = {}
        for feed_range in feed_ranges:
            if await container.is_feed_range_subset(feed_range, feed_range_from_pk):
                parent_feed_range = feed_range
                break
        # [END is_feed_range_subset]


        # Query a sorted list of items that were changed for one feed range.
        # The asynchronous client returns asynchronous iterators for its query methods;
        # as such, we iterate over it by using an async for loop
        # [START query_items_change_feed]
        async for queried_item in container.query_items_change_feed(feed_range=feed_ranges[0]):
            print(json.dumps(queried_item, indent=True))
        # [END query_items_change_feed]

        # Query a sorted list of items that were changed for one feed range from beginning.
        # The asynchronous client returns asynchronous iterators for its query methods;
        # as such, we iterate over it by using an async for loop
        # [START query_items_change_feed_from_beginning]
        async for queried_item in container.query_items_change_feed(feed_range=feed_ranges[0], start_time="Beginning"):
            print(json.dumps(queried_item, indent=True))
        # [END query_items_change_feed_from_beginning]

        await client.delete_database(database_name)
        print("Sample done running!")

if __name__ == "__main__":
    asyncio.run(examples_async())



# The MIT License (MIT)
# Copyright (c) Microsoft Corporation. All rights reserved.
import os
import unittest
import uuid

import pytest

import azure.cosmos.exceptions as exceptions
import test_config
import vector_test_data
from azure.cosmos import CosmosClient as CosmosSyncClient
from azure.cosmos import http_constants
from azure.cosmos.aio import CosmosClient
from azure.cosmos.partition_key import PartitionKey


def verify_ordering(item_list, distance_function):
    for i in range(len(item_list)):
        assert item_list[i]["text"] == vector_test_data.get_ordered_item_texts()[i]
    if distance_function == "euclidean":
        for i in range(len(item_list) - 1):
            assert item_list[i]["SimilarityScore"] <= item_list[i + 1]["SimilarityScore"]
    else:
        for i in range(len(item_list) - 1):
            assert item_list[i]["SimilarityScore"] >= item_list[i + 1]["SimilarityScore"]

@pytest.mark.cosmosSearchQuery
class TestVectorSimilarityQueryAsync(unittest.IsolatedAsyncioTestCase):
    """Test to check vector similarity queries behavior."""

    client: CosmosClient = None
    sync_client: CosmosSyncClient = None
    config = test_config.TestConfig
    host = config.host
    masterKey = config.masterKey
    connectionPolicy = config.connectionPolicy
    TEST_CONTAINER_ID = "Vector Similarity Container " + str(uuid.uuid4())

    @classmethod
    def setUpClass(cls):
        if (cls.masterKey == '[YOUR_KEY_HERE]' or
                cls.host == '[YOUR_ENDPOINT_HERE]'):
            raise Exception(
                "You must specify your Azure Cosmos account values for "
                "'masterKey' and 'host' at the top of this class to run the "
                "tests.")
        cls.sync_client = CosmosSyncClient(cls.host, cls.masterKey)
        cls.test_db = cls.sync_client.create_database(str(uuid.uuid4()))
        cls.created_quantized_cosine_container = cls.test_db.create_container(
            id="quantized" + cls.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_2_PARTITIONS,
            indexing_policy=test_config.get_vector_indexing_policy(embedding_type="quantizedFlat"),
            vector_embedding_policy=test_config.get_vector_embedding_policy(data_type="float32",
                                                                            distance_function="cosine",
                                                                            dimensions=128))
        cls.created_flat_euclidean_container = cls.test_db.create_container(
            id="flat" + cls.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_2_PARTITIONS,
            indexing_policy=test_config.get_vector_indexing_policy(embedding_type="flat"),
            vector_embedding_policy=test_config.get_vector_embedding_policy(data_type="float32",
                                                                            distance_function="euclidean",
                                                                            dimensions=128))
        cls.created_diskANN_dotproduct_container = cls.test_db.create_container(
            id="diskANN" + cls.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_2_PARTITIONS,
            indexing_policy=test_config.get_vector_indexing_policy(embedding_type="diskANN"),
            vector_embedding_policy=test_config.get_vector_embedding_policy(data_type="float32",
                                                                            distance_function="dotproduct",
                                                                            dimensions=128))
        cls.created_large_container = cls.test_db.create_container(
            id="large_container" + cls.TEST_CONTAINER_ID,
            partition_key=PartitionKey(path="/pk"),
            offer_throughput=test_config.TestConfig.THROUGHPUT_FOR_2_PARTITIONS,
            indexing_policy=test_config.get_vector_indexing_policy(embedding_type="quantizedFlat"),
            vector_embedding_policy=test_config.get_vector_embedding_policy(data_type="float32",
                                                                            distance_function="cosine",
                                                                            dimensions=2))
        for item in vector_test_data.get_vector_items():
            cls.created_quantized_cosine_container.create_item(item)
            cls.created_flat_euclidean_container.create_item(item)
            cls.created_diskANN_dotproduct_container.create_item(item)

    @classmethod
    def tearDownClass(cls):
        try:
            cls.sync_client.delete_database(cls.test_db.id)
        except exceptions.CosmosHttpResponseError:
            pass

    async def asyncSetUp(self):
        self.client = CosmosClient(self.host, self.masterKey)
        self.test_db = self.client.get_database_client(self.test_db.id)
        self.created_flat_euclidean_container = self.test_db.get_container_client(self.created_flat_euclidean_container.id)
        self.created_quantized_cosine_container = self.test_db.get_container_client(self.created_quantized_cosine_container.id)
        self.created_diskANN_dotproduct_container = self.test_db.get_container_client(self.created_diskANN_dotproduct_container.id)
        self.created_large_container = self.test_db.get_container_client(self.created_large_container.id)

    async def asyncTearDown(self):
        await self.client.close()

    async def test_wrong_vector_search_queries_async(self):
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")
        # try to send a vector search query without limit filters
        query = "SELECT c.text, VectorDistance(c.embedding, [{}]) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}])".format(vector_string, vector_string)
        try:
            [item async for item in self.created_large_container.query_items(query=query)]
            pytest.fail("Client should not allow queries without filters.")
        except ValueError as e:
            assert "Executing a vector search query without TOP or LIMIT can consume many RUs very fast and" \
                   " have long runtimes. Please ensure you are using one of the two filters with your" \
                   " vector search query." in e.args[0]

        # try to send a vector search query specifying the ordering as ASC or DESC
        query = "SELECT c.text, VectorDistance(c.embedding, [{}]) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}]) ASC".format(vector_string,
                                                                                               vector_string)
        try:
            [item async for item in self.created_large_container.query_items(query=query)]
            pytest.fail("Client should not allow queries with ASC/DESC.")
        except exceptions.CosmosHttpResponseError as e:
            assert e.status_code == http_constants.StatusCodes.BAD_REQUEST
            assert ("One of the input values is invalid." in e.message
                    or "Specifying a sorting order (ASC or DESC) with VectorDistance function is not supported." in e.message)

    async def test_vector_search_environment_variables_async(self):
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")
        query = "SELECT TOP 10 c.text, VectorDistance(c.embedding, [{}]) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}])".format(vector_string, vector_string)
        os.environ["AZURE_COSMOS_MAX_ITEM_BUFFER_VECTOR_SEARCH"] = "1"
        try:
            [item async for item in self.created_large_container.query_items(query=query)]
            pytest.fail("Config was not set correctly.")
        except ValueError as e:
            assert e.args[0] == ("Executing a vector search query with more items than the max is not allowed. "
                                 "Please ensure you are using a limit smaller than the max, or change the max.")

        os.environ["AZURE_COSMOS_MAX_ITEM_BUFFER_VECTOR_SEARCH"] = "50000"

        os.environ["AZURE_COSMOS_DISABLE_NON_STREAMING_ORDER_BY"] = "False"
        [item async for item in self.created_large_container.query_items(query=query)]

    async def test_ordering_distances_async(self):
        # Besides ordering distances, we also verify that the query text properly replaces any set embedding policies
        # load up previously calculated embedding for the given string
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")
        # test euclidean distance
        for i in range(1, 11):
            # we define queries with and without specs to directly use the embeddings in our container policies
            vanilla_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}]) AS " \
                            "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}])".format(str(i),
                                                                                                       vector_string,
                                                                                                       vector_string)
            specs_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'euclidean'}}) AS " \
                          "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'euclidean'}})" \
                .format(str(i), vector_string, vector_string)

            flat_list = [item async for item in self.created_flat_euclidean_container.query_items(query=vanilla_query)]
            verify_ordering(flat_list, "euclidean")

            quantized_list = [item async for item in self.created_quantized_cosine_container.query_items(query=specs_query)]
            verify_ordering(quantized_list, "euclidean")

            disk_ann_list = [item async for item in self.created_diskANN_dotproduct_container.query_items(query=specs_query)]
            verify_ordering(disk_ann_list, "euclidean")
        # test cosine distance
        for i in range(1, 11):
            vanilla_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}]) AS " \
                            "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}])".format(str(i),
                                                                                                       vector_string,
                                                                                                       vector_string)
            specs_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'cosine'}}) AS " \
                          "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'cosine'}})" \
                .format(str(i), vector_string, vector_string)

            flat_list = [item async for item in self.created_flat_euclidean_container.query_items(query=specs_query)]
            verify_ordering(flat_list, "cosine")

            quantized_list = [item async for item in self.created_quantized_cosine_container.query_items(query=vanilla_query)]
            verify_ordering(quantized_list, "cosine")

            disk_ann_list = [item async for item in self.created_diskANN_dotproduct_container.query_items(query=specs_query)]
            verify_ordering(disk_ann_list, "cosine")
        # test dot product distance
        for i in range(1, 11):
            vanilla_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}]) AS " \
                            "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}])".format(str(i),
                                                                                                       vector_string,
                                                                                                       vector_string)
            specs_query = "SELECT TOP {} c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'dotproduct'}}) AS " \
                          "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'dotproduct'}})" \
                .format(str(i), vector_string, vector_string)

            flat_list = [item async for item in self.created_flat_euclidean_container.query_items(query=specs_query)]
            verify_ordering(flat_list, "dotproduct")

            quantized_list = [item async for item in self.created_quantized_cosine_container.query_items(query=specs_query)]
            verify_ordering(quantized_list, "dotproduct")

            disk_ann_list = [item async for item in self.created_diskANN_dotproduct_container.query_items(query=vanilla_query)]
            verify_ordering(disk_ann_list, "dotproduct")

    async def test_vector_query_pagination_async(self):
        # load up previously calculated embedding for the given string
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")

        query = "SELECT TOP 8 c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'cosine'}}) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': " \
                "'cosine'}})".format(vector_string, vector_string)

        query_iterable = self.created_quantized_cosine_container.query_items(query=query,
                                                                             max_item_count=3)
        all_fetched_res = []
        count = 0
        item_pages = query_iterable.by_page()
        async for items in item_pages:
            count += 1
            all_fetched_res.extend([item async for item in items])
        assert count == 3
        assert len(all_fetched_res) == 8
        verify_ordering(all_fetched_res, "cosine")

    async def test_vector_query_large_data_async(self):
        # test different limit queries on a larger data set
        embedding_value = 0.0001
        for i in range(2000):
            item = {"id": str(i), "pk": i % 2, "embedding": [embedding_value, embedding_value]}
            await self.created_large_container.create_item(item)
            embedding_value += 0.0001

        query = "SELECT c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})" \
                " OFFSET 0 LIMIT 1000"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 1000

        query = "SELECT DISTINCT c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})" \
                " OFFSET 0 LIMIT 1000"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 1000

        query = "SELECT TOP 750 c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 750

        query = "SELECT DISTINCT TOP 750 c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 750

        query = "SELECT c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})" \
                " OFFSET 1000 LIMIT 500"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 500

        query = "SELECT DISTINCT c.id, VectorDistance(c.embedding, [0.0001, 0.0001], false," \
                " {'distanceFunction': 'cosine'}) AS SimilarityScore FROM c ORDER BY" \
                " VectorDistance(c.embedding, [0.0001, 0.0001], false, {'distanceFunction': 'cosine'})" \
                " OFFSET 1000 LIMIT 500"

        query_iterable = self.created_large_container.query_items(query=query)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 500

    async def test_vector_query_cross_partition_response_hook_async(self):
        # load up previously calculated embedding for the given string
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")

        query = "SELECT TOP 5 c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'cosine'}}) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': " \
                "'cosine'}})".format(vector_string, vector_string)

        response_hook = test_config.ResponseHookCaller()
        query_iterable = self.created_quantized_cosine_container.query_items(query=query,
                                                                             response_hook=response_hook)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 5
        assert response_hook.count == 2

    async def test_vector_query_partitioned_response_hook_async(self):
        # load up previously calculated embedding for the given string
        vector_string = vector_test_data.get_embedding_string("I am having a wonderful day.")

        query = "SELECT TOP 4 c.text, VectorDistance(c.embedding, [{}], false, {{'distanceFunction': 'cosine'}}) AS " \
                "SimilarityScore FROM c ORDER BY VectorDistance(c.embedding, [{}], false, {{'distanceFunction': " \
                "'cosine'}})".format(vector_string, vector_string)

        response_hook = test_config.ResponseHookCaller()
        query_iterable = self.created_quantized_cosine_container.query_items(query=query,
                                                                             partition_key='1',
                                                                             response_hook=response_hook)
        result_list = [item async for item in query_iterable]
        assert len(result_list) == 4
        assert response_hook.count == 1


if __name__ == "__main__":
    unittest.main()
